"""
Helper methods/classes for 'copr-backend-*' scripts.
"""

import os
import sys
import logging

from copr_common.background_worker import BackgroundWorker
from copr_backend.frontend import FrontendClient
from copr_backend.helpers import (BackendConfigReader, get_redis_logger)


class BackendBackgroundWorker(BackgroundWorker):
    """
    copr-backend-process-* abstraction
    """

    # This is still a base class, abstract methods will be overriden
    # in its descendants
    # pylint: disable=abstract-method

    def __init__(self):
        super().__init__()

        be_cfg = self.args.backend_config or '/etc/copr/copr-be.conf'
        self.opts = BackendConfigReader(be_cfg).read()

        self.frontend_client = FrontendClient(self.opts, self.log,
                                              try_indefinitely=True)

    def _switch_logger_to_redis(self):
        logger_name = '{}.{}.pid-{}'.format(
            sys.argv[0],
            'managed' if self.args.worker_id else 'manual',
            os.getpid(),
        )

        self.log = get_redis_logger(self.opts, logger_name,
                                    self.redis_logger_id)
        if not self.args.daemon and not self.args.silent:
            # when executing from commandline - on foreground - we want to
            # print something to stderr as well
            self.log.addHandler(logging.StreamHandler())

    def preparations_for_manager(self):
        # setup logging early, to have as complete logs as possible
        self._switch_logger_to_redis()



import os
import json
import logging
import tempfile
import shutil

import testlib
from testlib import assert_logs_exist, AsyncCreaterepoRequestFactory

from copr_common.redis_helpers import get_redis_connection
from copr_backend.createrepo import (
    BatchedCreaterepo,
    MAX_IN_BATCH,
)

from copr_backend.helpers import BackendConfigReader

# pylint: disable=attribute-defined-outside-init


class TestBatchedCreaterepo:
    def setup_method(self):
        self.workdir = tempfile.mkdtemp(prefix="copr-batched-cr-test-")
        self.config_file = testlib.minimal_be_config(self.workdir, {
            "redis_db": 9,
            "redis_port": 7777,
        })
        self.config = BackendConfigReader(self.config_file).read()
        self.redis = get_redis_connection(self.config)
        self.request_createrepo = AsyncCreaterepoRequestFactory(self.redis)
        self.redis.flushdb()
        self._pid = os.getpid()

    def teardown_method(self):
        shutil.rmtree(self.workdir)
        self.redis.flushdb()

    def _prep_batched_repo(self, some_dir, full=False, add=None, delete=None, rpms_to_remove=None):
        self.bcr = BatchedCreaterepo(
            some_dir,
            full,
            add if add is not None else ["subdir_add_1", "subdir_add_2"],
            delete if delete is not None else ["subdir_del_1", "subdir_del_2"],
            rpms_to_remove if rpms_to_remove is not None else [],
            logging.getLogger(),
            backend_opts=self.config,
        )
        return self.bcr

    def test_batched_createrepo_normal(self):
        some_dir = "/some/dir/name:pr:135"
        bcr = self._prep_batched_repo(some_dir)
        bcr.make_request()

        keys = self.redis.keys()
        assert len(keys) == 1
        assert keys[0].startswith("createrepo_batched::{}::".format(some_dir))
        redis_dict = self.redis.hgetall(keys[0])
        redis_task = json.loads(redis_dict["task"])
        assert len(redis_dict) == 1
        assert redis_task == {
            "appstream": True,
            "devel": False,
            "add": ["subdir_add_1", "subdir_add_2"],
            "delete": ["subdir_del_1", "subdir_del_2"],
            "full": False,
            "rpms_to_remove": [],
        }
        self.request_createrepo.get(some_dir)
        # appstream=False has no effect, others beat it, appstream=False
        # makes it non-matching.
        self.request_createrepo.get(some_dir, {"add": ["add_2"], "appstream": False})
        self.request_createrepo.get(some_dir, {"add": [], "delete": ["del_1"]})
        self.request_createrepo.get(some_dir, {"add": [], "delete": ["del_2"]})
        assert not bcr.check_processed()
        assert bcr.options() == (False,
                                 set(["add_1"]),
                                 set(["del_1", "del_2"]),
                                 set([]))
        assert len(bcr.notify_keys) == 3

        our_key = keys[0]

        bcr.commit()
        keys = self.redis.keys()
        count_non_finished = 0
        for key in keys:
            assert key != our_key
            task_dict = self.redis.hgetall(key)
            if "status" in task_dict:
                assert task_dict["status"] == "success"
            else:
                count_non_finished += 1
        assert count_non_finished == 1


    def test_batched_createrepo_already_done(self):
        some_dir = "/some/dir/name"
        bcr = self._prep_batched_repo(some_dir)
        key = bcr.make_request()
        self.request_createrepo.get(some_dir)
        self.redis.hset(key, "status", "success")
        assert bcr.check_processed()
        assert self.redis.hgetall(key) == {}
        assert bcr.notify_keys == []  # nothing to commit()

    def test_batched_createrepo_other_already_done(self, caplog):
        some_dir = "/some/dir/name:pr:3"
        bcr = self._prep_batched_repo(some_dir)
        key = bcr.make_request()

        # create two other requests, one is not to be processed
        self.request_createrepo.get(some_dir)
        self.request_createrepo.get(some_dir, {"add": ["add_2"]}, done=True)

        # nobody processed us
        assert not bcr.check_processed()

        # we only process the first other request
        assert bcr.options() == (False, set(["add_1"]), set(), set())
        assert len(bcr.notify_keys) == 1  # still one to notify
        assert self.redis.hgetall(key) == {}
        assert len(caplog.record_tuples) == 2
        assert_logs_exist("already processed, skip", caplog)

    def test_batched_createrepo_devel_mismatch(self, caplog):
        some_dir = "/some/dir/name:pr:5"
        bcr = self._prep_batched_repo(some_dir)
        key = bcr.make_request()

        # create two other requests, one is not to be processed
        self.request_createrepo.get(some_dir, {"add": ["add_2"], "devel": True})
        self.request_createrepo.get(some_dir)

        # nobody processed us
        assert not bcr.check_processed()

        # we only process the first other request
        assert bcr.options() == (False, set(["add_1"]), set(), set())
        assert len(bcr.notify_keys) == 1  # still one to notify
        assert self.redis.hgetall(key) == {}
        assert len(caplog.record_tuples) == 2
        assert_logs_exist("'devel' attribute doesn't match", caplog)

    def test_batched_createrepo_full_we_take_others(self):
        some_dir = "/some/dir/name:pr:take_others"
        bcr = self._prep_batched_repo(some_dir, full=True, add=[], delete=[])
        key = bcr.make_request()

        task = self.redis.hgetall(key)
        task_json = json.loads(task["task"])
        assert task_json["full"]
        assert task_json["add"] == [] == task_json["delete"]

        # create three other requests, one is not to be processed
        self.request_createrepo.get(some_dir, {"add": ["add_2"], "devel": True})
        self.request_createrepo.get(some_dir)
        self.request_createrepo.get(some_dir, {"add": [], "delete": ["del_1"]})

        assert len(self.redis.keys()) == 4
        assert not bcr.check_processed()
        assert len(self.redis.keys()) == 3

        assert bcr.options() == (False, {"add_1"}, {"del_1"}, set())
        assert len(bcr.notify_keys) == 2

    def test_batched_createrepo_full_others_take_us(self):
        some_dir = "/some/dir/name:pr:others_take_us"
        bcr = self._prep_batched_repo(some_dir)
        key = bcr.make_request()
        task = self.redis.hgetall(key)
        task_json = json.loads(task["task"])
        assert not task_json["full"]
        assert not bcr.check_processed()

        # create four other requests, one is not to be processed
        self.request_createrepo.get(some_dir, {"add": ["add_2"]})
        self.request_createrepo.get(some_dir, done=True)
        self.request_createrepo.get(some_dir, {"add": [], "delete": [], "full": True})
        self.request_createrepo.get(some_dir, {"add": [], "delete": ["del_1"]})

        assert bcr.options() == (True, set(), {"del_1"}, set())
        assert len(bcr.notify_keys) == 3

    def test_batched_createrepo_task_limit(self, caplog):
        some_dir = "/some/dir/name:pr:limit"

        # request a createrepo run (devel == False!)
        bcr = self._prep_batched_repo(some_dir)
        key = bcr.make_request()
        assert len(self.redis.keys()) == 1

        # add 'add_1' task
        self.request_createrepo.get(some_dir)

        # another background, not considered because devel == True
        self.request_createrepo.get(some_dir, {"add": ["add_2"], "devel": True})

        # Fill-up the MAX_IN_BATCH quota with background requests, and add
        # two more.
        for i in range(3, 3 + MAX_IN_BATCH):
            add_dir = "add_{}".format(i)
            self.request_createrepo.get(some_dir, {"add": [add_dir]})

        # MAX_IN_BATCH + 2 more above + one is ours
        assert len(self.redis.keys()) == MAX_IN_BATCH + 2 + 1

        # Nobody processed us, drop us from DB
        assert not bcr.check_processed()
        assert len(self.redis.keys()) == MAX_IN_BATCH + 2

        # What directories should be processed at once?  Drop add_2 as it is
        # devel=True.
        expected = {"add_{}".format(i) for i in range(1, MAX_IN_BATCH + 3)}
        expected.remove("add_2")
        assert len(expected) == MAX_IN_BATCH + 1

        full, add, remove, rpms_to_remove = bcr.options()
        assert (full, remove, rpms_to_remove) == (False, set(), set())
        # check that the batch is this request + (MAX_IN_BATCH - 1)
        assert len(add) == MAX_IN_BATCH - 1

        # The redis.keys() list isn't sorted, and even if it was - our own
        # PID in the key would make the final order.  Therefore we don't know
        # which items are skipped, but we know there are two left for the next
        # batch.
        assert len(expected-add) == 2

        # Nothing unexpected should go here.
        assert add-expected == set()

        assert len(bcr.notify_keys) == MAX_IN_BATCH - 1
        assert self.redis.hgetall(key) == {}

        log_entries_expected = {
            "Batch copr-repo limit",
            "Has already a status? False",
        }

        msg_count = len(caplog.record_tuples)
        if msg_count == 3:
            # we have one miss in roughly MAX_IN_BATCH tasks in the redis
            # database, so there's roughly 1:MAX_IN_BATCH chance this message
            # doesn't appear.
            log_entries_expected.add("'devel' attribute doesn't match")
        else:
            assert msg_count == 2

        assert_logs_exist(log_entries_expected, caplog)

        bcr.commit()
        without_status = set()
        for key in self.redis.keys():
            if not self.redis.hget(key, "status"):
                data = json.loads(self.redis.hget(key, "task"))
                for add_dir in data["add"]:
                    without_status.add(add_dir)
        assert "add_2" in without_status
        assert len(without_status) == 3



""" test counting priority of build task """

import pytest

from copr_backend.rpm_builds import BuildQueueTask, PRIORITY_SECTION_SIZE
from copr_backend.daemons.build_dispatcher import _PriorityCounter


def test_priority_numbers():
    prio = _PriorityCounter()
    assert prio.get_priority(BuildQueueTask({
        "build_id": "7",
        "task_id": "7",
        "project_owner": "cecil",
    })) == 1
    assert prio.get_priority(BuildQueueTask({
        "build_id": "8",
        "task_id": "8",
        "project_owner": "cecil",
    })) == 2
    assert prio.get_priority(BuildQueueTask({
        "build_id": "88",
        "task_id": "88",
        "project_owner": "cecil",
        "background": True,
    })) == 1  # background jobs have separate counters

    assert prio.get_priority(BuildQueueTask({
        "build_id": "9",
        "task_id": "9-fedora-rawhide-x86_64",
        "chroot": "fedora-rawhide-x86_64",
        "project_owner": "cecil",
        "background": True,
    })) == 1
    assert prio.get_priority(BuildQueueTask({
        "build_id": "10",
        "task_id": "10-fedora-rawhide-i386",
        "chroot": "fedora-rawhide-i386",
        "project_owner": "cecil",
        "background": True,
    })) == 2
    assert prio.get_priority(BuildQueueTask({
        "build_id": "10",
        "task_id": "10-fedora-rawhide-aarch64",
        "chroot": "fedora-rawhide-aarch64",
        "project_owner": "cecil",
        "background": True,
    })) == 1
    assert prio.get_priority(BuildQueueTask({
        "build_id": "11",
        "task_id": "11-fedora-rawhide-aarch64",
        "chroot": "fedora-rawhide-aarch64",
        "project_owner": "bedrich",
        "background": True,
    })) == 1

@pytest.mark.parametrize('background,result',
                         [(True, 2*PRIORITY_SECTION_SIZE), (False, 0)])
def test_frontend_priority(background, result):
    task = BuildQueueTask({
        "build_id": "9",
        "task_id": "9",
        "project_owner": "cecil",
        "background": background,
    })
    assert task.frontend_priority == result
    task = BuildQueueTask({
        "build_id": "9",
        "task_id": "9-fedora-rawhide-x86_64",
        "project_owner": "cecil",
        "background": background,
    })
    assert task.frontend_priority == result + PRIORITY_SECTION_SIZE


def test_sandbox_priority():
    prio = _PriorityCounter()
    assert prio.get_priority(BuildQueueTask({
        "build_id": "9",
        "task_id": "9",
        "project_owner": "cecil",
        "background": False,
        "sandbox": "cecil/foo--submitter",
    })) == 1

    assert prio.get_priority(BuildQueueTask({
        "build_id": "9",
        "task_id": "9-fedora-rawhide-x86_64",
        "chroot": "fedora-rawhide-x86_64",
        "project_owner": "cecil",
        "background": True,
        "sandbox": "cecil/foo--submitter",
    })) == 1  # different arch

    assert prio.get_priority(BuildQueueTask({
        "build_id": "10",
        "task_id": "10-fedora-rawhide-x86_64",
        "chroot": "fedora-rawhide-x86_64",
        "project_owner": "cecil",
        "background": True,
        "sandbox": "cecil/foo--submitter",
    })) == 2  # the same arch

    assert prio.get_priority(BuildQueueTask({
        "build_id": "11",
        "task_id": "11-fedora-rawhide-x86_64",
        "chroot": "fedora-rawhide-x86_64",
        "project_owner": "cecil",
        "background": True,
        "sandbox": "cecil/baz--submitter",
    })) == 1  # the same arch, but different sandbox



"""
example of cancellable thread task
"""

import threading
import logging


def _stderr_logger():
    logging.basicConfig(level=logging.DEBUG)
    return logging.getLogger('root')


class CancellableThreadTask:
    """
    Start ``method`` in background thread, and wait for external "cancel" event
    (when ``cb_check_canceled`` returns True).  When the "cancel" event happens,
    call the ``cb_cancel`` calback.
    """

    # pylint: disable=too-many-arguments,too-few-public-methods
    def __init__(self, method, cb_check_canceled, cb_cancel,
                 check_period=5, log=None):
        self.method = method
        self.check = cb_check_canceled
        self.cancel = cb_cancel
        self.result = None
        self.check_period = check_period
        self.log = log or _stderr_logger()

    def _background_run_wrapper(self, call, result, *args, **kwargs):
        try:
            result.result = call(*args, **kwargs)
        except Exception:  # pylint: disable=broad-except
            # No exceptions to avoid de-synchronization of the threads
            self.log.exception("Exception during cancellable method")

    def run(self, *args, **kwargs):
        """ execute the self.method with args/kwargs """
        self.log.debug("starting background method: %s", self.method)
        thread = threading.Thread(
            target=self._background_run_wrapper,
            args=(self.method, self) + args,
            kwargs=kwargs,
        )
        thread.start()
        while True:
            self.log.debug("checking liveness")
            thread.join(timeout=self.check_period)
            if not thread.is_alive():
                self.log.debug("thread ended")
                break

            if self.check(*args, **kwargs):
                self.log.info("calling cancel callback, and waiting")
                try:
                    self.cancel(*args, **kwargs)
                except Exception:  # pylint: disable=broad-except
                    # No exceptions to avoid de-synchronization of the threads
                    self.log.exception("Exception during cancel request")
                thread.join()
                break

        return self.result



# coding: utf-8

import os
import json
import logging
import tempfile
import shutil
from munch import Munch

from copr_common.tree import walk_limited
from copr_common.redis_helpers import get_redis_connection
from copr_backend.background_worker_build import BackendError
from copr_backend.helpers import (
    copy2_but_hardlink_rpms,
    get_chroot_arch,
    get_redis_logger,
    format_filename,
)
from copr_backend.constants import LOG_REDIS_FIFO

"""
SOME TESTS REQUIRES RUNNING REDIS
"""

MODULE_REF = "copr_backend.helpers"


class TestHelpers(object):

    def setup_method(self, method):
        self.opts = Munch(
            redis_db=9,
            redis_port=7777,
        )

        self.rc = get_redis_connection(self.opts)
        # remove leftovers from previous tests
        self.rc.delete(LOG_REDIS_FIFO)

    def teardown_method(self, method):
        pass

    def test_redis_logger_exception(self):
        log = get_redis_logger(self.opts, "copr_backend.test", "backend")
        try:
            raise BackendError("foobar")
        except Exception as err:
            log.exception("error occurred: %s", err)

        (_, raw_message) = self.rc.blpop([LOG_REDIS_FIFO])
        data = json.loads(raw_message)
        assert data.get("who") == "backend"
        assert data.get("levelno") == logging.ERROR
        assert "error occurred: Backend process error: foobar\n" in data["msg"]
        assert 'raise BackendError("foobar")' in data["msg"]

    def test_get_chroot_arch(self):
        assert get_chroot_arch("fedora-26-x86_64") == "x86_64"
        assert get_chroot_arch("epel-7-ppc64le") == "ppc64le"
        assert get_chroot_arch("epel-7-ppc64") == "ppc64"

    def test_format_filename(self):
        assert format_filename("ed", "1.14.2", "5.fc30", "", "x86_64") == "ed-1.14.2-5.fc30.x86_64"
        assert format_filename("ed", "1.14.2", "5.fc30", "", "x86_64", zero_epoch=True) == "ed-0:1.14.2-5.fc30.x86_64"
        assert format_filename("ed", "1.14.2", "5.fc30", "2", "x86_64") == "ed-2:1.14.2-5.fc30.x86_64"

        split = ("ed", "1.14.2", "5.fc30", "2", "x86_64")
        assert format_filename(zero_epoch=True, *split) == "ed-2:1.14.2-5.fc30.x86_64"

    def test_walk_limited(self):
        paths = [
            "user1/foo/srpm-builds/111/foo-1.src.rpm",
            "user1/foo/srpm-builds/111/foo-1.log",
            "user1/foo/srpm-builds/222/foo-2.log",
            "user1/foo/fedora-rawhide-x86_64/222-foo/foo-2.something.rpm",
            "user1/foo/fedora-rawhide-x86_64/note.txt",
            "user2/bar/.disable-appstream",
        ]
        with tempfile.TemporaryDirectory(prefix="copr-test-walk") as results:
            for path in paths:
                path = os.path.join(results, path)
                os.makedirs(os.path.dirname(path), exist_ok=True)
                open(path, "a").close()

            # The unlimited walk should be equal to built-in walk function
            assert list(os.walk(results)) == list(walk_limited(results))

            # Our fake directory strucutre isn't ten levels deep
            assert not list(walk_limited(results, mindepth=10))

            # A corner case with zero maxdepth
            output = list(walk_limited(results, maxdepth=0))
            root, subdirs, files = output[0]
            assert root == results
            assert set(subdirs) == {"user2", "user1"}
            assert files == []

            # Combination of both mindepth and maxdepth
            output = sorted(list(walk_limited(results, mindepth=3, maxdepth=3)))
            root, subdirs, files = output[0]
            assert root == os.path.join(results, "user1/foo/fedora-rawhide-x86_64")
            assert subdirs == ["222-foo"]
            assert files == ["note.txt"]

            root, subdirs, files = output[1]
            assert root == os.path.join(results, "user1/foo/srpm-builds")
            assert set(subdirs) == {"222", "111"}
            assert files == []


    def test_recursive_copy_and_link_rpms(self):

        def _write(filename, contents):
            with open(filename, "w", encoding="utf-8") as fd:
                fd.write(contents)

        def _read(filename):
            with open(filename, "r", encoding="utf-8") as fd:
                return fd.read()

        with tempfile.TemporaryDirectory(prefix="copr-test-walk") as workdir:
            src = os.path.join(workdir, "src")
            dst = os.path.join(workdir, "dst")
            os.mkdir(src)

            subdir = os.path.join(src, "subdir")
            subdir_dst = os.path.join(dst, "subdir")
            os.mkdir(subdir)

            textfile = os.path.join(subdir, "test.txt")
            textfile_dst = os.path.join(subdir_dst, "test.txt")
            _write(textfile, "text")

            rpmfile = os.path.join(subdir, "test.rpm")
            rpmfile_dst = os.path.join(subdir_dst, "test.rpm")
            _write(rpmfile, "rpmfile")

            shutil.copytree(src, dst, copy_function=copy2_but_hardlink_rpms)

            _write(textfile, "text changed")
            _write(rpmfile, "rpmfile re-signed")

            # touch the source files
            os.path.join(dst, "text")
            os.path.join(dst, "rpmfile")

            # linked file is affected
            assert _read(rpmfile_dst) == "rpmfile re-signed"
            # copied file is not affected
            assert _read(textfile_dst) == "text"



" test vm_alloc.py "

from unittest import mock

import pytest

from copr_backend.vm_alloc import (
    RemoteHostAllocationTerminated,
    ResallocHostFactory,
)

@mock.patch('copr_backend.vm_alloc.ResallocConnection')
def test_ticket(_rcon):
    hf = ResallocHostFactory()
    host = hf.get_host()

    host.ticket.closed = False
    host.ticket.ready = False
    assert not host.check_ready()
    assert host.hostname is None

    host.ticket.output = "1.1.1.1"
    host.ticket.ready = True
    assert host.check_ready()
    assert host.hostname == "1.1.1.1"

    host.ticket.output = (
        "---\n"
        "host: 1.2.3.4\n"
        "name: copr_pool_123456\n"
    )
    assert host.check_ready()
    assert host.hostname == "1.2.3.4"
    assert host.name == "copr_pool_123456"

    host.ticket.closed = True
    with pytest.raises(RemoteHostAllocationTerminated):
        assert host.check_ready()
    host.release()

@mock.patch('copr_backend.vm_alloc.ResallocConnection')
def test_ticket_with_args(rcon):
    hf = ResallocHostFactory()
    host = hf.get_host(sandbox='somesb', tags=['arch_x86_64'])
    host.ticket.closed = False
    host.ticket.ready = True
    host.ticket.output = "1.1.1.1"
    host.wait_ready()
    assert rcon.return_value.newTicket.call_args_list == [
        mock.call(['copr_builder', 'arch_x86_64'], 'somesb'),
    ]
    assert host.hostname == "1.1.1.1"

class _collect_side_effect:
    def __init__(self, results, host):
        self.returns = results
        self.counter = 0
        self.host = host
    def __call__(self):
        ret = self.returns[self.counter]
        self.host.ticket.closed = ret == 'closed'
        self.host.ticket.ready = ret is True
        self.host.ticket.output = "host" if ret else None
        self.counter += 1

@mock.patch('copr_backend.vm_alloc.time.sleep')
@mock.patch('copr_backend.vm_alloc.ResallocConnection')
def test_ticket_wait_ready_normal(_rcon, sleep):
    hf = ResallocHostFactory()
    host = hf.get_host()
    host.ticket.collect.side_effect = _collect_side_effect(
        [False, False, False, False, True],
        host,
    )
    expected_calls = [
        mock.call(3),
        mock.call(3),
        mock.call(6),
        mock.call(6),
    ]
    host.wait_ready()
    assert sleep.call_args_list == expected_calls
    assert host.hostname == "host"
    # cached
    host.wait_ready()
    assert sleep.call_args_list == expected_calls

@mock.patch('copr_backend.vm_alloc.time.sleep')
@mock.patch('copr_backend.vm_alloc.ResallocConnection')
def test_ticket_wait_ready_raises(_rcon, sleep):
    hf = ResallocHostFactory()
    host = hf.get_host()
    host.ticket.collect.side_effect = _collect_side_effect(
        [False, 'closed'],
        host,
    )
    expected_calls = [mock.call(3)]
    assert not host.wait_ready()
    assert sleep.call_args_list == expected_calls


@mock.patch('copr_backend.vm_alloc.time.sleep')
@mock.patch('copr_backend.vm_alloc.ResallocConnection')
def test_ticket_wait_ready_fallback(_rcon, sleep):
    hf = ResallocHostFactory()
    host = hf.get_host()

    host.ticket.collect.side_effect = _collect_side_effect(
        [False for _ in range(20)] + [True],
        host,
    )
    host.wait_ready()
    assert len(sleep.call_args_list) == 20



import json
import logging
import logging.handlers
import optparse
import os
import sys
import errno
import time
import types
import glob
import shlex
import shutil

import configparser
from configparser import ConfigParser

from operator import methodcaller

import traceback

import datetime
from threading import Thread

import subprocess

import munch
from munch import Munch

from copr_common.redis_helpers import get_redis_connection
from copr.v3 import Client
from copr_backend.constants import DEF_BUILD_USER, DEF_BUILD_TIMEOUT, DEF_CONSECUTIVE_FAILURE_THRESHOLD, \
    CONSECUTIVE_FAILURE_REDIS_KEY, default_log_format
from copr_backend.exceptions import CoprBackendError

from . import constants

DOMAIN = "fedorahosted.org"

LOG_COMPONENTS = [
    "build_dispatcher", "action_dispatcher",
    "backend", "actions", "worker", "modifyrepo", "pruner", "analyze-results",
]


def pyconffile(filename):
    """
    Load python file as configuration file, inspired by python-flask
    "from_pyfile()
    """
    d = types.ModuleType('bus_config({0})'.format(filename))
    d.__file__ = filename
    try:
        with open(filename) as config_file:
            exec(compile(config_file.read(), filename, 'exec'), d.__dict__)
    except IOError as e:
        e.strerror = 'Unable to load configuration file (%s)' % e.strerror
        raise
    return d


def cmd_debug(cmd, rc, out, err, log):
    log.info("cmd: %s", cmd)
    log.info("rc: %s", rc)
    log.info("stdout: %s", out)
    log.info("stderr: %s", err)


class CommandException(Exception):
    """ Raised upon run_cmd(..., check=True) failure """


def run_cmd(cmd, shell=False, timeout=None, logger=None, catch_timeout=False,
            check=False):
    """
    Runs given command in a subprocess, logs what is happening, return important
    info about the process.

    Params
    ------
    cmd: list(str) or str if shell==True
        command to be executed and its arguments
    shell: bool
        if the command should be interpreted by shell
    timeout:
        passed down to Popen.communicate().  See catch_timeout, too.
    logger:
        A logging.getLogger()-like object to use for dumping a useful info.  The
        method is log-silent by default.
    catch_timeout:
        When set to True, we catch timeout exceptions and log them, and return
        the status code 124.
    check:
        When set to True, raise CommandException() in case of CMD's non-zero
        exit status.

    Returns
    -------
    munch.Munch(stdout, stderr, returncode)
        executed cmd, standard output, error output, and the return code
    """

    start = time.time()

    def _info(logger, *args, **kwargs):
        if not logger:
            return
        logger.info(*args, **kwargs)

    process = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=shell, encoding="utf-8")

    pid = process.pid
    str_cmd = cmd
    if isinstance(cmd, list):
        str_cmd = " ".join([shlex.quote(a) for a in cmd])

    _info(logger, "Running command '%s' as PID %s", str_cmd, pid)

    timeouted = False
    try:
        (stdout, stderr) = process.communicate(timeout=timeout)
        code = process.returncode
    except subprocess.TimeoutExpired:
        _info(logger, "Timeouted (%s)", str_cmd)
        if not catch_timeout:
            raise
        # https://docs.python.org/3/library/subprocess.html#subprocess.Popen.communicate
        process.kill()
        stdout, stderr = process.communicate()

        code = 124  # inspired by /bin/timeout
        timeouted = True

    took_seconds = int(time.time() - start)

    outputs_dumped = f"\nstdout:\n{stdout}\nstderr:\n{stderr}\n"

    _info(
        logger,
        "Finished after %d seconds with exit code %s (%s)%s",
        took_seconds, code, str_cmd, outputs_dumped if code else "",
    )

    if check and code != 0:
        exc_msg = f"Command '{str_cmd}' failed, exit code '{code}"
        if not logger:
            # this has not yet been logged, make it a part of the exception
            exc_msg += outputs_dumped
        raise CommandException(exc_msg)

    return munch.Munch(
        cmd=cmd,
        stdout=stdout,
        stderr=stderr,
        timeouted=timeouted,
        returncode=code,
        pid=pid,
        str_cmd=str_cmd,
    )


def wait_log(log, reason="I don't know why.", timeout=5):
    """
    We need to wait a while, this should happen only when copr converges to
    boot-up/restart/..
    """
    if not log:
        return
    log.warning("I'm waiting %ss because: %s", timeout, reason)
    time.sleep(timeout)


class SortedOptParser(optparse.OptionParser):
    """Optparser which sorts the options by opt before outputting --help"""

    def format_help(self, formatter=None):
        self.option_list.sort(key=methodcaller("get_opt_string"))
        return optparse.OptionParser.format_help(self)


def _get_conf(cp, section, option, default, mode=None):
    """
    To make returning items from config parser less irritating

    :param mode: convert obtained value, possible modes:
      - None (default): do nothing
      - "bool" or "boolean"
      - "int"
      - "float"
    """
    # pylint: disable=too-many-return-statements

    if cp.has_section(section) and cp.has_option(section, option):
        if mode is None:
            return cp.get(section, option)
        elif mode in ["bool", "boolean"]:
            return cp.getboolean(section, option)
        elif mode == "int":
            return cp.getint(section, option)
        elif mode == "float":
            return cp.getfloat(section, option)
        elif mode == "list":
            return [x.strip() for x in cp.get(section, option).split(",")]
        elif mode == "path":
            path = cp.get(section, option)
            if path.startswith("~"):
                path = os.path.expanduser(path)
            path = os.path.abspath(path)
            path = os.path.normpath(path)

            return path
    return default


def _get_limits_conf(parser):
    err1 = ("Unexpected format of 'builds_max_workers_{0}' configuration "
            "option.  Please use format: "
            "builds_max_workers_{0} = {1}1=COUNT,{1}2=COUNT")
    err2 = ("Duplicate left value '{}' in 'builds_max_workers_{}' configuration")
    limits = {"arch": {}, "tag": {}, "arch_per_owner": {}}

    for config_type in ["arch", "tag", "arch_per_owner"]:
        option = "builds_max_workers_{}".format(config_type)
        raw = _get_conf(parser, "backend", option, None)
        if raw:
            raw_configs = raw.split(',')
            for limit_spec in raw_configs:
                try:
                    key, count = limit_spec.split("=")
                    key = key.strip()
                    count = int(count.strip())
                    if not key or not count:
                        raise CoprBackendError("Empty '{}' spec".format(option))
                    if key in limits[config_type]:
                        raise CoprBackendError(err2.format(key, config_type))
                    limits[config_type][key] = count
                except ValueError as orig:
                    new = CoprBackendError(err1.format(config_type,
                                                       config_type.upper()))
                    raise new from orig
    limits['sandbox'] = _get_conf(
        parser, "backend", "builds_max_workers_sandbox", 10, mode="int")
    limits['owner'] = _get_conf(
        parser, "backend", "builds_max_workers_owner", 20, mode="int")
    limits['userssh'] = _get_conf(
        parser, "backend", "builds_max_userssh", 2, mode="int")
    limits['blocked_owners'] = _get_conf(
        parser, "backend", "blocked_owners", [], mode="list")
    return limits


class BackendConfigReader(object):
    def __init__(self, config_file=None, ext_opts=None):
        self.config_file = config_file or "/etc/copr/copr-be.conf"
        self.ext_opts = ext_opts

    def read(self):
        try:
            opts = self._read_unsafe()
            if self.ext_opts:
                for key, value in self.ext_opts.items():
                    setattr(opts, key, value)

            if not opts.destdir:
                raise CoprBackendError(
                    "Incomplete Config - must specify"
                    " destdir in configuration")

            return opts

        except configparser.Error as e:
            raise CoprBackendError(
                "Error parsing config file: {0}: {1}".format(
                    self.config_file, e))

    def _read_unsafe(self):
        cp = ConfigParser()
        cp.read(self.config_file)

        opts = Munch()

        opts.results_baseurl = _get_conf(
            cp, "backend", "results_baseurl", "http://copr-be")

        opts.frontend_base_url = _get_conf(
            cp, "backend", "frontend_base_url", "http://copr-fe")

        opts.frontend_auth = _get_conf(
            cp, "backend", "frontend_auth", "PASSWORDHERE")

        opts.redis_host = _get_conf(
            cp, "backend", "redis_host", "127.0.0.1")

        opts.redis_port = _get_conf(
            cp, "backend", "redis_port", "6379")

        opts.redis_password = _get_conf(
            cp, "backend", "redis_password", None)

        opts.redis_db = _get_conf(
            cp, "backend", "redis_db", "0")

        opts.do_sign = _get_conf(
            cp, "backend", "do_sign", False, mode="bool")

        opts.keygen_host = _get_conf(
            cp, "backend", "keygen_host", "copr-keygen.cloud.fedoraproject.org")

        opts.build_user = _get_conf(
            cp, "backend", "build_user", DEF_BUILD_USER)

        opts.build_groups_count = _get_conf(
            cp, "backend", "build_groups", 1, mode="int")

        opts.sign_domain = _get_conf(
            cp, "backend", "sign_domain", DOMAIN)

        opts.build_groups = []
        for group_id in range(opts.build_groups_count):
            archs = _get_conf(cp, "backend",
                              "group{0}_archs".format(group_id),
                              default="i386,x86_64").split(",")
            group = {
                "id": int(group_id),
                "name": _get_conf(cp, "backend", "group{0}_name".format(group_id), "PC"),
                "archs": archs,
                "spawn_playbook": _get_conf(
                    cp, "backend", "group{0}_spawn_playbook".format(group_id),
                    default="/srv/copr-work/provision/builderpb-PC.yml"),
                "terminate_playbook": _get_conf(
                    cp, "backend", "group{0}_terminate_playbook".format(group_id),
                    default="/srv/copr-work/provision/terminatepb-PC.yml"),
                "max_vm_total": _get_conf(
                    cp, "backend", "group{}_max_vm_total".format(group_id),
                    # default=16, mode="int"),
                    default=8, mode="int"),
                "max_vm_per_user": _get_conf(
                    cp, "backend", "group{}_max_vm_per_user".format(group_id),
                    default=4, mode="int"),
                "max_spawn_processes": _get_conf(
                    cp, "backend", "group{}_max_spawn_processes".format(group_id),
                    default=2, mode="int"),
                "vm_spawn_min_interval": _get_conf(
                    cp, "backend", "group{}_vm_spawn_min_interval".format(group_id),
                    default=30, mode="int"),
                "vm_dirty_terminating_timeout": _get_conf(
                    cp, "backend", "group{}_vm_dirty_terminating_timeout".format(group_id),
                    default=120, mode="int"),
                "vm_health_check_period": _get_conf(
                    cp, "backend", "group{}_vm_health_check_period".format(group_id),
                    default=120, mode="int"),
                "vm_health_check_max_time": _get_conf(
                    cp, "backend", "group{}_vm_health_check_max_time".format(group_id),
                    default=300, mode="int"),
                "vm_max_check_fails": _get_conf(
                    cp, "backend", "group{}_vm_max_check_fails".format(group_id),
                    default=2, mode="int"),
                "vm_terminating_timeout": _get_conf(
                    cp, "backend", "group{}_vm_terminating_timeout".format(group_id),
                    default=600, mode="int"),
                "playbook_timeout": _get_conf(
                    cp, "backend", "group{}_playbook_timeout".format(group_id),
                    default=None, mode="int"),
            }
            opts.build_groups.append(group)

        opts.vm_cycle_timeout = _get_conf(
            cp, "backend", "vm_cycle_timeout",
            default=10, mode="int")
        opts.vm_ssh_check_timeout = _get_conf(
            cp, "backend", "vm_ssh_check_timeout",
            default=5, mode="int")

        opts.destdir = _get_conf(cp, "backend", "destdir", None, mode="path")

        opts.fedmsg_enabled = _get_conf(
            cp, "backend", "fedmsg_enabled", False, mode="bool")
        opts.sleeptime = _get_conf(
            cp, "backend", "sleeptime", 5, mode="int")
        opts.timeout = _get_conf(
            cp, "builder", "timeout", DEF_BUILD_TIMEOUT, mode="int")
        opts.consecutive_failure_threshold = _get_conf(
            cp, "builder", "consecutive_failure_threshold",
            DEF_CONSECUTIVE_FAILURE_THRESHOLD, mode="int")

        opts.resalloc_connection = _get_conf(
            cp, "backend", "resalloc_connection", "http://localhost:49100")
        opts.builds_max_workers = _get_conf(
            cp, "backend", "builds_max_workers",
            default=60, mode="int")
        opts.builds_limits = _get_limits_conf(cp)

        opts.actions_max_workers = _get_conf(
            cp, "backend", "actions_max_workers",
            default=10, mode="int")

        opts.prune_workers = _get_conf(
            cp, "backend", "prune_workers",
            default=None, mode="int")

        opts.log_dir = _get_conf(
            cp, "backend", "log_dir", "/var/log/copr-backend/")
        opts.log_level = _get_conf(
            cp, "backend", "log_level", "info")
        opts.log_format = _get_conf(
            cp, "backend", "log_format", default_log_format)

        opts.statsdir = _get_conf(
            cp, "backend", "statsdir", "/var/lib/copr/public_html/stats")

        opts.stats_templates_dir = _get_conf(
            cp, "backend", "stats_templates_dir",
            os.path.join(os.path.dirname(__file__), "stats_templates"))

        opts.prune_days = _get_conf(cp, "backend", "prune_days", None, mode="int")

        opts.gently_gpg_sha256 = _get_conf(
            cp, "backend", "gently_gpg_sha256", True, mode="bool")

        opts.aws_cloudfront_distribution = _get_conf(
            cp, "backend", "aws_cloudfront_distribution", None)

        opts.pulp_content_url = _get_conf(
            cp, "backend", "pulp_content_url", None)

        # ssh options
        opts.ssh = Munch()
        opts.ssh.builder_config = _get_conf(
            cp, "ssh", "builder_config", "/home/copr/.ssh/builder_config")

        opts.msg_buses = []
        for bus_config in glob.glob('/etc/copr/msgbuses/*.conf'):
            opts.msg_buses.append(pyconffile(bus_config))

        # thoughts for later
        # ssh key for connecting to builders?
        # cloud key stuff?
        #
        return opts


def get_project_info(front_url, username, projectname):
    """
    Get JSON data from Frontend API page with the info about the given project.
    """
    client = Client({"copr_url": front_url})
    return client.project_proxy.get(username, projectname)


def uses_devel_repo(front_url, username, projectname, project=None):
    """
    Detect if the given project uses the "pre-release" builds being managed by
    metadata in a seprate devel/ directory.  If project object is given, we
    use it instead of doing additional API call to frontend to get the info.
    """
    if project is None:
        client = Client({"copr_url": front_url})
        project = client.project_proxy.get(username, projectname)
    return bool(project.get("devel_mode", False))


def register_build_result(opts=None, failed=False):
    """
    Remember fails to redis.
    Successful build resets counter to zero.

    :param opts: BackendConfig, when opts not provided default config location will be used
    :param boolean failed: failure flag
    :param str origin: name of component produced failure, default: `builder`
    """
    if opts is None:
        opts = BackendConfigReader().read()

    conn = get_redis_connection(opts)

    key = CONSECUTIVE_FAILURE_REDIS_KEY
    if not failed:
        conn.set(key, 0)
    else:
        conn.incr(key)


def format_tb(ex, ex_traceback):
    tb_lines = traceback.format_exception(ex.__class__, ex, ex_traceback)
    return ''.join(tb_lines)


class RedisPublishHandler(logging.Handler):
    """
    :type rc: StrictRedis
    """
    def __init__(self, rc, who, level=logging.NOTSET,):
        super(RedisPublishHandler, self).__init__(level)

        self.rc = rc
        self.who = who

    def emit(self, record):
        # copr specific semantics

        # Alternative to copy.deepcopy().  If we edit the original record
        # object, any other following log handler would get the modified
        # variant.
        record = logging.makeLogRecord(record.__dict__)

        record.who = self.who

        # First argument to 'log.exception()' should be 'str' type.  If it is
        # not, we need to convert it before using '%' operator below.
        record.msg = str(record.msg)

        # For the message arguments, it is better to expand them right now
        # instead of relying on method in json.dumps(..., default=default)
        # and even worse rely on it's reverse action in RedisLogHandler.
        record.msg = record.msg % record.args
        if record.exc_info:
            _, error, tb = record.exc_info
            record.msg += "\n" + format_tb(error, tb)

        # cleanup the hard to json.dumps() stuff
        record.exc_info = None
        record.exc_text = None
        record.args = ()

        try:
            self.rc.rpush(constants.LOG_REDIS_FIFO, json.dumps(record.__dict__))
        # pylint: disable=W0703
        except Exception as error:
            _, _, ex_tb = sys.exc_info()
            sys.stderr.write("Failed to publish log record to redis, {}"
                             .format(format_tb(error, ex_tb)))

def get_redis_log_handler(opts, component):
    """
    Get RedisPublishHandler object sending the events to Redis.

    :param opts: BackendConfigReader.read() output dict.
    :param name: Name of the logger, e.g. 'root'.
    :param component: Where to log the file in /var/log/copr-backend.  When
        the component is 'pruner', the file will be 'pruner.log'.
    """
    assert component in LOG_COMPONENTS
    rc = get_redis_connection(opts)
    # level=DEBUG, by default we send everything logger gives us
    handler = RedisPublishHandler(rc, component, level=logging.DEBUG)
    return handler


def get_redis_logger(opts, name, who):
    logger = logging.getLogger(name)
    level = getattr(opts, 'log_level', 'debug')
    level = getattr(logging, level.upper(), logging.DEBUG)
    logger.setLevel(level)

    if not logger.handlers:
        logger.addHandler(get_redis_log_handler(opts, who))

    return logger


def create_file_logger(name, filepath, fmt=None):
    logger = logging.getLogger(name)

    if not logger.handlers:
        handler = logging.handlers.WatchedFileHandler(filename=filepath)
        handler.setFormatter(fmt if fmt is not None else default_log_format)
        logger.addHandler(handler)

    return logger


def utc_now():
    """
    :return datetime.datetime: Current utc datetime with specified timezone
    """
    u = datetime.datetime.now(datetime.UTC)
    return u


def silent_remove(filename):
    try:
        os.remove(filename)
    except OSError as e: # this would be "except OSError, e:" before Python 2.6
        if e.errno != errno.ENOENT: # errno.ENOENT = no such file or directory
            raise # re-raise exception if a different error occurred


def get_backend_opts():
    args = sys.argv[1:]
    parser = optparse.OptionParser("\ncopr-be [options]")
    parser.add_option("-c", "--config", default="/etc/copr/copr-be.conf",
                      dest="config_file",
                      help="config file to use for copr-be run")

    opts, args = parser.parse_args(args)
    if not os.path.exists(opts.config_file):
        sys.stderr.write("No config file found at: {0}\n".format(
            opts.config_file))
        sys.exit(1)

    config_file = os.path.abspath(opts.config_file)
    config_reader = BackendConfigReader(config_file, {})
    return config_reader.read()


def ensure_dir_exists(path, log):
    """
    Create the directory if it doesn't exist yet.  But still "raise" if it can
    not be created.
    """
    try:
        os.makedirs(path)
        log.info("Directory %s created", path)
    except FileExistsError:
        if os.path.isdir(path):
            return
        log.exception("File %s exists but not a directory", path)
        raise
    except:
        log.exception("Can't create %s directory", path)
        raise


def get_chroot_arch(chroot):
    return chroot.rsplit("-", 2)[2]


def format_evr(epoch, version, release):
    """
    Return evr in format (epoch:)version-release.  The argument 'epoch' should
    be integer value or null (but we rather also consider "strings" values).
    """
    if epoch is not None:
        if isinstance(epoch, int) or epoch.isdigit():
            return f"{epoch}:{version}-{release}"
    return f"{version}-{release}"


def format_filename(name, version, release, epoch, arch, zero_epoch=False):
    if not epoch.isdigit() and zero_epoch:
        epoch = "0"
    if epoch.isdigit():
        return "{}-{}:{}-{}.{}".format(name, epoch, version, release, arch)
    return "{}-{}-{}.{}".format(name, version, release, arch)


def call_copr_repo(directory, rpms_to_remove=None, devel=False, add=None, delete=None, timeout=None,
                   logger=None, appstream=True, do_stat=False):
    """
    Execute 'copr-repo' tool, and return True if the command succeeded.
    """
    cmd = ["copr-repo", "--batched", directory]
    def opt_multiply(option, subdirs):
        args = []
        if not subdirs:
            return []
        for subdir in subdirs:
            if subdir is None:
                # this should never happen, but better to skip
                # this than kill some backend process
                continue
            args += [option, subdir]
        return args
    cmd += opt_multiply('--add', add)
    cmd += opt_multiply('--delete', delete)
    cmd += opt_multiply('--rpms-to-remove', rpms_to_remove)
    if not appstream:
        cmd += ['--no-appstream-metadata']
    if devel:
        cmd += ['--devel']

    if do_stat:
        cmd += ["--do-stat"]

    result = run_cmd(cmd, timeout=timeout, logger=logger, catch_timeout=True)
    if result.returncode and logger:
        logger.error("Createrepo failed")

    return not result.returncode

def build_target_dir(build_id, package_name=None):
    build_id = int(build_id)
    if not package_name:
        return "{:08d}".format(build_id)
    return "{:08d}-{}".format(build_id, package_name)

def build_chroot_log_name(build_id, package_name=None):
    return 'build-{}.log'.format(build_target_dir(build_id, package_name))


def copy2_but_hardlink_rpms(src, dest, **kwargs):
    """
    Link RPMs while otherwise copying using the shutil.copytree() method.
    Use it like:
        shutil.copytree(src, dst,
                        copy_function=copy2_but_hardlink_rpms())
    """
    if src.endswith(".rpm"):
        return os.link(src, dest)
    # This is per help(shutil.copytree), copy2 is used by default.
    return shutil.copy2(src, dest, **kwargs)



"""
BuildBackgroundWorker class + internals.
"""

import functools
import glob
import logging
import os
import shutil
import statistics
import time
import json
import shlex

from datetime import datetime
from packaging import version
from cachetools.func import ttl_cache

from copr_common.enums import StatusEnum, StorageEnum
from copr_common.helpers import (
    USER_SSH_DEFAULT_EXPIRATION,
    USER_SSH_MAX_EXPIRATION,
    USER_SSH_EXPIRATION_PATH,
)

from copr_backend.background_worker import BackendBackgroundWorker
from copr_backend.cancellable_thread import CancellableThreadTask
from copr_backend.constants import build_log_format
from copr_backend.exceptions import (
    CoprSignError,
    CoprBackendError,
    FrontendClientException,
)
from copr_backend.helpers import (
    run_cmd, register_build_result, format_evr,
)
from copr_backend.job import BuildJob
from copr_backend.msgbus import MessageSender
from copr_backend.sign import sign_rpms_in_dir, get_pubkey
from copr_backend.sshcmd import SSHConnection, SSHConnectionError
from copr_backend.vm_alloc import ResallocHostFactory
from copr_backend.storage import storage_for_job


MAX_HOST_ATTEMPTS = 3
MAX_SSH_ATTEMPTS = 5
MIN_BUILDER_VERSION = "1.3.1"
CANCEL_CHECK_PERIOD = 5
DATETIME_FORMAT = "%Y-%m-%d %H:%M"

MESSAGES = {
    "give_up_repo":
        "Giving up waiting for {0} repository, "
        "please try to manually regenerate the DNF repository "
        "(e.g. by 'copr-cli regenerate-repos <project_name>')",
    "repo_waiting":
        "Waiting for %s repository",
    "copr_rpmbuild_missing":
        "The copr-rpmbuild package was not found: {}",
}

COMMANDS = {
    "rpm_q_builder": "rpm -q copr-rpmbuild --qf \"%{VERSION}\n\"",
    "echo_authorized_keys": "echo {0} >> /root/.ssh/authorized_keys",
    "set_expiration": "echo -n {0} > " + USER_SSH_EXPIRATION_PATH,
    "cat_expiration": "cat {0}".format(USER_SSH_EXPIRATION_PATH),
}


class BuildRetry(Exception):
    """
    Stop processing the build on the current host, and ask for a new
    one (even though the same may be given by the Resalloc server).
    So we re-try the build as long as the host seems to reply on SSH
    channel and it makes at least some sense.  We also retry on at most
    MAX_HOST_ATTEMPTS hosts when the hosts suddenly become unusable.
    """

class BuildCanceled(Exception):
    """ Synchronous cancel request received, fail the build! """
    def __str__(self):
        return "Build was canceled"

class BackendError(Exception):
    """ Generic build failure. """
    def __str__(self):
        return "Backend process error: {}".format(super().__str__())


def _average_step(values):
    """
    Calculate average step between ``values``.  It's expected that
    ``values`` contains at least two items.
    """
    if len(values) < 4:
        return float("inf")
    previous = None
    intervals = []
    for value in values:
        if previous:
            intervals.append(value - previous)
        previous = value
    return statistics.mean(intervals)


class LoggingPrivateFilter(logging.Filter):
    """
    Filter-out messages that can potentially reveal some data that
    should stay private.
    """
    def filter(self, record):
        if record.exc_info:
            traceback = record.exc_info[2]
            while traceback.tb_next:
                traceback = traceback.tb_next
            fname = traceback.tb_frame.f_code.co_filename
            fline = traceback.tb_frame.f_lineno

            record.exc_info = None
            record.exc_text = ""
            record.msg = record.msg + " (in {}:{})".format(
                fname, fline,
            )

        return 1


def skipped_for_source_build(f):
    """ Mark method no-op for the source build (srpm-builds dir) """
    @functools.wraps(f)
    def wrapper(self, *args, **kwargs):
        if self.job.chroot == 'srpm-builds':
            self.log.debug("Skipping method %s for source build", f.__name__)
            return None
        return f(self, *args, **kwargs)
    return wrapper


class BuildBackgroundWorker(BackendBackgroundWorker):
    """
    The (S)RPM build logic.
    """
    # pylint: disable=too-many-instance-attributes

    redis_logger_id = 'worker'

    def __init__(self):
        super().__init__()
        self.sender = None
        self.builder_dir = "/var/lib/copr-rpmbuild"
        self.builder_livelog = os.path.join(self.builder_dir, "main.log")
        self.builder_results = os.path.join(self.builder_dir, "results")
        self.ssh = None
        self.root_ssh = None
        self.job = None
        self.host = None
        self.canceled = False
        self.last_hostname = None
        self.storage = None

    @classmethod
    def adjust_arg_parser(cls, parser):
        parser.add_argument(
            "--build-id",
            type=int,
            required=True,
            help="build ID to process",
        )
        parser.add_argument(
            "--chroot",
            required=True,
            help="chroot name (or 'srpm-builds')",
        )

    @property
    def name(self):
        """ just a name for logging purposes """
        return "backend.worker-{}".format(self.worker_id)

    def _prepare_result_directory(self, job):
        """
        Create backup directory and move there results from previous build.
        """
        try:
            os.makedirs(job.results_dir)
        except FileExistsError:
            pass

        results_dir_entries = list(os.scandir(job.results_dir))
        if not results_dir_entries:
            return

        backup_dir_name = "prev_build_backup"
        backup_dir = os.path.join(job.results_dir, backup_dir_name)
        self.log.info("Cleaning target directory, results from previous build storing in %s",
                      backup_dir)

        if not os.path.exists(backup_dir):
            os.makedirs(backup_dir)

        files = (x.name for x in results_dir_entries if x.name != backup_dir_name)
        for filename in files:
            file_path = os.path.join(job.results_dir, filename)
            if os.path.isfile(file_path):
                if file_path.endswith((".info", ".log", ".log.gz")):
                    os.rename(file_path, os.path.join(backup_dir, filename))

                elif not file_path.endswith(".rpm"):
                    os.remove(file_path)
            else:
                shutil.rmtree(file_path)

    def _setup_resultdir_and_logging(self):
        """ Prepare the result directory and log file ASAP """
        self._prepare_result_directory(self.job)
        handler = logging.FileHandler(
            filename=self.job.backend_log,
        )
        handler.setLevel(logging.INFO)
        handler.setFormatter(build_log_format)
        handler.addFilter(LoggingPrivateFilter())
        self.log.addHandler(handler)

    def _mark_starting(self):
        """
        Announce to the frontend that the build is starting. Frontend may reject
        build to start.
        """
        self.log.info("Marking build as starting")
        self.job.status = StatusEnum("starting")
        if not self.frontend_client.starting_build(self.job.to_dict()):
            raise BackendError("Frontend forbade to start the job {}".format(
                self.job.task_id))

    def _check_copr_builder(self):
        rc, out, err = self.ssh.run_expensive(COMMANDS["rpm_q_builder"])
        if rc != 0:
            raise BuildRetry(MESSAGES["copr_rpmbuild_missing"].format(err))
        if version.parse(out) < version.parse(MIN_BUILDER_VERSION):
            # retry for this issue indefinitely, till the VM is removed and
            # up2date is spawned
            raise BuildRetry("Minimum version for builder is {}"
                             .format(MIN_BUILDER_VERSION))

    def _check_vm(self):
        """
        Check that the VM is OK to start the build
        """
        self.log.info("Checking that builder machine is OK")
        self._check_copr_builder()

        # The output won't be live and will appear only after this command
        # finishes. Making it live is nontrivial but we have a good code for
        # doing so in `resallocserver.manager.run_command`. Praiskup plans to
        # generalize it into a separate package that we could eventually use
        # here.
        cmd = "copr-builder-ready " + self.job.chroot
        self.log.info("Running remote command: %s", cmd)
        rc, stdout, stderr = self.ssh.run_expensive(
            cmd, subprocess_timeout=660)
        self.log.info(stdout)
        if rc:
            self.log.info(stderr)
            raise BuildRetry("Builder wasn't ready, trying a new one")

    def _fill_build_info_file(self):
        """
        Places "build.info" which contains job build_id and worker IP
        into the directory with downloaded files.
        """
        info_file_path = os.path.join(self.job.results_dir, "build.info")
        self.log.info("Filling build.info file with builder info")
        try:
            with open(info_file_path, 'w') as info_file:
                info_file.writelines([
                    "build_id={}".format(self.job.build_id),
                    "\nbuilder_ip={}".format(self.host.hostname)])

        except Exception as error:
            raise BackendError("Can't write to {}: {}".format(
                info_file_path, error,
            ))

    def _update_frontend_task(self, data):
        self.log.info("Sending build state back to frontend: %s",
                      json.dumps(data, indent=4))
        self.frontend_client.update(data)

    def _mark_running(self, attempt):
        """
        Announce everywhere that a build process started now.
        """
        self._proctitle("Job {}, host info: {}".format(self.job.task_id,
                                                       self.host.info))
        self.job.started_on = time.time()
        self.job.status = StatusEnum("running")

        if attempt > 0:
            # TODO: invent new message type for re-try
            self.log.info("Not re-notifying FE and msg buses for the new host.")
            return

        data = {"builds": [self.job.to_dict()]}
        self._update_frontend_task(data)

        for topic in ['build.start', 'chroot.start']:
            self.sender.announce(topic, self.job, self.last_hostname)

    def _mark_finished(self):
        self.job.ended_on = time.time()

        # At this point, NEVER want to re-try the build by subsequent
        # BackgroundWorker process.  Let's enforce "finished" state.
        allowed_states = "failed", "succeeded"
        if self.job.status not in [StatusEnum(s) for s in allowed_states]:
            self.log.warning("Switching not-finished job state to 'failed'")
            self.job.status = StatusEnum("failed")

        text_status = StatusEnum(self.job.status)
        self.log.info("Worker %s build, took %s", text_status,
                      self.job.took_seconds)
        data = {"builds": [self.job.to_dict()]}
        self._update_frontend_task(data)
        self.sender.announce("build.end", self.job, self.last_hostname)

    def _parse_results(self):
        """
        Parse `results.json` and update the `self.job` object.
        """
        # When user SSH is allowed, we don't download any results from the
        # builder for safety reasons. Don't try to parse anything.
        if self.job.ssh_public_keys:
            return

        path = os.path.join(self.job.results_dir, "results.json")
        if not os.path.exists(path):
            raise BackendError("results.json file not found in resultdir")
        with open(path, "r") as f:
            results = json.load(f)
        self.job.results = results

    def _wait_for_repo(self, repo_id, baseurl):
        """
        Wait a while for initial createrepo, and eventually fail the build
        if the waiting is not successful.
        """
        waiting_since = time.time()
        while time.time() - waiting_since < 60:
            exists = self.storage.repository_exists(
                self.job.project_dirname,
                self.job.chroot,
                baseurl,
            )
            if exists:
                return

            # Either (a) the very first copr-repo run in this chroot dir
            # is still running on background (or failed), or (b) we are
            # hitting the race condition between
            # 'rm -rf repodata && mv .repodata repodata' sequence that
            # is done in createrepo_c.  Try again after some time.
            self.log.info(MESSAGES["repo_waiting"], repo_id)
            time.sleep(2)

        # This should never happen, but if yes - we need to debug
        # properly.  Give up waiting, and fail the build.  That should
        # motivate people to report bugs.
        raise BackendError(MESSAGES["give_up_repo"].format(repo_id))

    def _wait_for_repos(self):
        """
        Wait a while for initial createrepo of `copr_base` and `copr_coprdir`,
        and eventually fail the build if the waiting is not successful.
        """
        if self.job.chroot == 'srpm-builds':
            # we don't need copr_base repodata for srpm builds
            return

        for repo in self.job.repos:
            # If these don't exist, it doesn't make sense to start the build.
            # It is also our fault. If user-defined repos doesn't exist, we
            # don't care here and we'd rather let the build fail while DNF
            # trying to enable them.
            if repo["id"] not in ["copr_base", "copr_coprdir"]:
                continue
            self._wait_for_repo(repo["id"], repo["baseurl"])

    def _get_build_job(self):
        """
        Per self.args, obtain BuildJob instance.
        """
        if self.args.chroot == "srpm-builds":
            target = "get-srpm-build-task/{}".format(self.args.build_id)
        else:
            target = "get-build-task/{}-{}".format(self.args.build_id,
                                                   self.args.chroot)

        try:
            resp = self.frontend_client.get(target)
        except FrontendClientException as ex:
            self.log.error("Failed to download build info: %s", str(ex))
            msg = "Failed to get the build task {}".format(target)
            raise BackendError(msg) from ex

        self.job = BuildJob(resp.json(), self.opts)
        self.job.started_on = time.time()
        if not self.job.chroot:
            raise BackendError("Frontend job doesn't provide chroot")
        self.storage = storage_for_job(self.job, self.opts, self.log)

    def _drop_host(self):
        """
        Deallocate assigned host.  We can call this multiple times in row (to
        make sure the host is deallocated), so this needs to stay idempotent.
        """
        if not self.host:
            return

        self.log.info("Releasing VM back to pool")
        self.host.release()
        self.host = None

    def _proctitle(self, text):
        text = "Builder for task {}: {}".format(self.job.task_id, text)
        self.log.debug("setting title: %s", text)
        self.setproctitle(text)

    def _check_build_interrupted(self):
        """
        Should we interrupt a running worker?
        """
        return (self._check_failed_resalloc_ticket()
                or self._cancel_task_check_request()
                or self._build_timeouted())

    @ttl_cache(ttl=10*60)
    def _check_failed_resalloc_ticket(self):
        """
        Did the resalloc ticket fail?
        """
        self.host.ticket.collect()
        self.log.info("Periodic builder liveness probe: %s",
                      "dead" if self.host.ticket.failed else "alive")
        return self.host.ticket.failed

    def _build_timeouted(self):
        """
        When build timeouts, it should be handled by `copr-rpmbuild` and the
        builder machine should mark itself as finished. When it fails to do so,
        we have this fail-safe to know if a build timeouted and should be
        terminated.
        """
        # Wait some time (1 hour) after the configured timeout for the builder
        # to terminate itself.
        timestamp = self.job.started_on + self.job.timeout + 60 * 60
        limit = datetime.fromtimestamp(timestamp)
        return datetime.now() > limit

    def _cancel_task_check_request(self):
        """
        Was the build canceled by the user?
        """
        self.log.info("Checking for cancel request")
        self.canceled = bool(self.redis_get_worker_flag("cancel_request"))
        return self.canceled

    def _cancel_if_requested(self):
        """
        Raise BuildCanceled exception if there's already a request for
        cancellation.  This is useful as "quick and cheap check" before starting
        some expensive task that would have to be later canceled anyways.
        We can call this multiple times, anytime we feel it is appropriate.
        """
        if self._cancel_task_check_request():
            self.log.warning("Canceling the build early")
            self._drop_host()
            raise BuildCanceled

    def _cancel_vm_allocation(self):
        self.redis_set_worker_flag("canceling", 1)
        self._drop_host()

    def _alloc_host(self):
        """
        Set self.host with ready RemoteHost, and return True.  Keep re-trying
        upon allocation failure.  Return False if the request was canceled.
        """

        self.log.info("VM allocation process starts")
        vm_factory = ResallocHostFactory(server=self.opts.resalloc_connection)
        while True:
            self.host = vm_factory.get_host(self.job.tags, self.job.sandbox)
            self.log.info("Trying to allocate VM: %s", self.host.info)
            self._proctitle("Waiting for VM, info: {}".format(self.host.info))
            success = CancellableThreadTask(
                self.host.wait_ready,
                self._cancel_task_check_request,
                self._cancel_vm_allocation,
                check_period=CANCEL_CHECK_PERIOD,
                log=self.log,
            ).run()
            if self.canceled:
                raise BuildCanceled
            if success:
                self.log.info("Allocated host %s", self.host.info)
                self.last_hostname = self.host.hostname
                return
            time.sleep(60)
            self.log.error("VM allocation failed, trying to allocate new VM")

    def _alloc_ssh_connection(self):
        self.log.info("Allocating ssh connection to builder")
        self.ssh = SSHConnection(
            user=self.opts.build_user,
            port=self.host.ssh_port,
            host=self.host.hostname,
            config_file=self.opts.ssh.builder_config,
            log=self.log,
        )

    def _discard_running_worker(self):
        """
        This is "canceling" callback to CancellableThreadTask, so please never
        raise any exception.  The worst case scenario is that nothing is
        canceled.
        """
        self._proctitle("Canceling running task...")
        self.redis_set_worker_flag("canceling", 1)
        try:
            cmd = "copr-rpmbuild-cancel"
            rc, out, err = self.ssh.run_expensive(cmd, max_retries=3)
            if rc:
                self.log.warning("Can't cancel build\nout:\n%s\nerr:\n%s",
                                 out, err)
                return
            self.log.info("Cancel request succeeded\nout:\n%serr:\n%s",
                          out, err)
        except SSHConnectionError:
            self.log.error("Can't ssh to cancel build.")

    def _start_remote_build(self):
        """ start the RPM build on builder on background """
        command = "copr-rpmbuild --verbose --drop-resultdir"
        if self.job.chroot == "srpm-builds":
            command += " --srpm --task-url {task_url} --detached"
        else:
            command += " --task-url {task_url} --chroot {chroot} --detached"
        command = command.format(task_url=self.job.task_url,
                                 chroot=self.job.chroot)

        self.log.info("Starting remote build: %s", command)
        rc, stdout, stderr = self.ssh.run_expensive(command)
        if rc:
            raise BackendError("Can't start copr-rpmbuild,\nout:\n{}err:\n{}"
                               .format(stdout, stderr))
        self.log.info("The copr-rpmbuild seems started, per:\nstdout: %s\n"
                      "stderr: %s\n", stdout, stderr)

    def _tail_log_file(self):
        """ Return None if OK, or failure reason as str """
        live_cmd = "copr-rpmbuild-log"
        with open(self.job.builder_log, 'w') as logfile:
            # We can not use 'max_retries' here because that would concatenate
            # the attempts to the same log file.
            if self.ssh.run(live_cmd, stdout=logfile, stderr=logfile,
                            subprocess_timeout=None):
                return "{} shouldn't exit != 0".format(live_cmd)
        return None

    def _retry_for_ssh_failures(self, method, *args, **kwargs):
        """
        Retry running the ``method`` indefinitely when SSHConnectionError occurs
        more frequently than each 2 minutes.
        """
        attempt = 0
        ssh_failures = []
        while True:
            attempt += 1
            try:
                self.log.info("Downloading the builder-live.log file, "
                              "attempt %s", attempt)
                return method(*args, **kwargs)
            except SSHConnectionError as exc:
                ssh_failures += [time.time()]
                if _average_step(ssh_failures[-4:]) < 120:
                    self.log.error("Giving up for unstable SSH, failures: %s",
                                   ", ".join([str(x) for x in ssh_failures]))
                    raise
                sleep = 10
                self.log.warning("SSH connection lost on #%s attempt, "
                                 "let's retry after %ss, %s", attempt, sleep, exc)
                time.sleep(sleep)
                continue

    def _transfer_log_file(self):
        """
        Since the tail process can be "watched" for a very long time, there's
        quite some chance we loose ssh connection in the meantime.  Therefore
        re-try downloading it till the ssh "looks" to be working.

        This is "cancellable" task, so we should NEVER RAISE any exception.
        """
        try:
            return self._retry_for_ssh_failures(self._tail_log_file)
        except SSHConnectionError as exc:
            return "Stopped following builder for broken SSH: {}".format(exc)

    def _compress_logs(self):
        """
        Compress builder-live.log, backend.log, and fedora-review.log using gzip.
        Never raise any exception!
        """
        logs = [
            self.job.builder_log,
            self.job.backend_log,
            self.job.review_log,
        ]

        # For automatic redirect from log to log.gz, consider configuring
        # Lighttpd like:
        #
        #   url.redirect += ( "^/(.*)/redirect-builder-live.log$" => "/$1/builder-live.log.gz" )
        #   url.rewrite-if-not-file = ("^/(.*)/builder-live.log$" => "/$1/redirect-builder-live.log")
        #   url.redirect += ( "^/(.*)/redirect-backend.log$" => "/$1/backend.log.gz" )
        #   url.rewrite-if-not-file += ("^/(.*)/backend.log$" => "/$1/redirect-backend.log")
        #   url.redirect += ( "^/(.*)/redirect-backend.log$" => "/$1/fedora-review.log.gz" )
        #   url.rewrite-if-not-file += ("^/(.*)/backend.log$" => "/$1/redirect-fedora-review.log")
        #
        #   $HTTP["url"] =~ "\.log\.gz$" {
        #       magnet.attract-physical-path-to = ( "/etc/lighttpd/content-encoding-gzip-if-exists.lua" )
        #       mimetype.assign = ("" => "text/plain" )
        #   }
        #
        # .. here the Lua script looks just like:
        #
        #   if (lighty.stat(lighty.env["physical.path"])) then
        #     lighty.header["Content-Encoding"] = "gzip"
        #   end
        #
        # Or Apache with:
        #     <FilesMatch "^(builder-live|backend|fedora-review)\.log$">
        #     RewriteEngine on
        #     RewriteCond %{REQUEST_FILENAME} !-f
        #     RewriteRule ^(.*)$ %{REQUEST_URI}.gz [R]
        #     </FilesMatch>

        for src in logs:
            dest = src + ".gz"
            if os.path.exists(dest):
                # This shouldn't ever happen, but if it happened - gzip below
                # would interactively ask whether we want to overwrite the
                # existing file, and it would deadlock the worker.
                self.log.error("Compressed log %s exists", dest)
                continue

            if not os.path.exists(src) and src == self.job.review_log:
                # fedora-review.log has a good chance of not existing
                # We should be ready for other similar files
                self.log.warning("Not trying to compress %s as it does not exist", src)
                continue

            self.log.info("Compressing %s by gzip", src)
            res = run_cmd(["gzip", src], logger=self.log)
            if res.returncode not in [0, 2]:
                self.log.error("Unable to compress file %s", src)

    def _download_results(self):
        """
        Retry rsync-download the results several times.
        """
        filter_ = None
        if self.job.ssh_public_keys:
            self.log.info("Builder allowed user SSH, not downloading the "
                          "results for safety reasons.")
            filter_ = ["+ success", "+ *.spec", "- *"]

        self.log.info("Downloading results from builder")
        self.ssh.rsync_download(
            self.builder_results + "/",
            self.job.results_dir,
            logfile=self.job.rsync_log_name,
            max_retries=2,
            filter_=filter_,
        )

    def _upload_results_to_storage(self):
        """
        Upload build results to an appropriate storage. Duplicate the data,
        don't remove them from the original place. At the end, a `self._cleanup`
        method will remove the temporary data.

        Currently, we need to run several more methods between this method and
        the cleanup but it should be possible to rearrange the steps, so that
        this method could upload the results and remove the temporary files
        at the same time.
        """
        rpms = self.storage.find_build_results(self.job.results_dir)
        result = self.storage.upload_build_results(
            rpms,
            self.job.chroot,
            build_id=self.job.build_id,
        )

        # Only PulpStorage returns package HREFs
        if not result:
            return

        hrefs = [x["pulp_href"] for x in result.values()]
        success = self.storage.create_repository_version(
            self.job.project_dirname,
            self.job.chroot,
            hrefs,
        )
        if not success:
            raise BackendError(
                "Pulp: Failed to create repository version for {0} and {1}"
                .format(self.job.project_dirname, self.job.chroot)
            )

    def _check_build_success(self):
        """
        Raise BackendError if builder claims that the build failed.
        """
        self.log.info("Searching for 'success' file in resultdir")
        successfile = os.path.join(self.job.results_dir, "success")
        if not os.path.exists(successfile):
            raise BackendError("No success file => build failure")

    @skipped_for_source_build
    def _sign_built_packages(self):
        """
            Sign built rpms
             using `copr_username` and `copr_projectname` from self.job
             by means of obs-sign. If user builds doesn't have a key pair
             at sign service, it would be created through ``copr-keygen``

        :param chroot_dir: Directory with rpms to be signed
        :param pkg: path to the source package

        """
        self.log.info("Going to sign pkgs from source: %s in chroot: %s",
                      self.job.task_id, self.job.chroot_dir)

        sign_rpms_in_dir(
            self.job.project_owner,
            self.job.project_name,
            os.path.join(self.job.chroot_dir, self.job.target_dir_name),
            self.job.chroot,
            opts=self.opts,
            log=self.log
        )

        self.log.info("Sign done")

    def _do_createrepo(self):
        if self.job.chroot == 'srpm-builds':
            return

        kwargs = {
            "chroot_dir": self.job.chroot_dir,
            "target_dir_name": self.job.target_dir_name,
        }
        if not self.storage.publish_repository(self.job.chroot, **kwargs):
            raise BackendError("createrepo failed")

    def _get_srpm_build_details(self, job):
        build_details = {'srpm_url': ''}
        self.log.info("Retrieving SRPM info from %s", job.results_dir)

        # pylint: disable=unsubscriptable-object
        assert isinstance(self.job.results, dict)

        build_details["pkg_name"] = self.job.results["name"]
        build_details["pkg_version"] = format_evr(
            self.job.results["epoch"],
            self.job.results["version"],
            self.job.results["release"],
        )

        pattern = os.path.join(job.results_dir, '*.src.rpm')
        srpm_file = glob.glob(pattern)[0]
        srpm_name = os.path.basename(srpm_file)
        srpm_url = os.path.join(job.results_dir_url, srpm_name)
        build_details['srpm_url'] = srpm_url
        self.log.info("SRPM URL: %s", srpm_url)
        return build_details

    def _collect_built_packages(self, job):
        """
        Return all built RPM packages as one string, e.g.
        'copr-builder 0.68\ncopr-distgit-client 0.68\ncopr-rpmbuild 0.68'
        """
        self.log.info("Listing built binary packages in %s", job.results_dir)

        if self.job.ssh_public_keys:
            return ""

        # pylint: disable=unsubscriptable-object
        assert isinstance(self.job.results, dict)

        packages = []
        for pkg in self.job.results["packages"]:
            if pkg["arch"] == "src":
                continue
            packages.append("{0} {1}".format(pkg["name"], pkg["version"]))

        self.log.info("Built packages:\n%s", packages)
        return "\n".join(packages)


    def _get_build_details(self, job):
        """
        :return: dict with build_details
        :raises BackendError: Something happened with build itself
        """
        self.log.info("Getting build details")
        try:
            if job.chroot == "srpm-builds":
                build_details = self._get_srpm_build_details(job)
            else:
                build_details = {
                    "built_packages": self._collect_built_packages(job),
                }
            self.log.info("build details: %s", build_details)
        except Exception as e:
            self.log.exception("Can't collect build results for %s",
                               job.task_id)
            raise BackendError("Can not deduct build details") from e

        return build_details

    @skipped_for_source_build
    def _add_pubkey(self):
        """
        Adds pubkey.gpg with public key to ``chroot_dir`` using
        ``copr_username`` and ``copr_projectname`` from self.job.
        """
        if not self.opts.do_sign:
            return

        self.log.info("Retrieving pubkey")

        # TODO: sign repodata as well ?
        user = self.job.project_owner
        project = self.job.project_name
        pubkey_path = os.path.join(self.job.destdir, "pubkey.gpg")

        # TODO: uncomment this when key revoke/change will be implemented
        # if os.path.exists(pubkey_path):
        #    return
        get_pubkey(user, project, self.log, self.opts.sign_domain, pubkey_path)
        self.log.info("Added pubkey for user %s project %s into: %s",
                      user, project, pubkey_path)

    @skipped_for_source_build
    def _setup_for_user_ssh(self):
        """
        Setup the builder for user SSH
        https://github.com/fedora-copr/debate/tree/main/user-ssh-builders

        If the builder setup for user SSH becomes more complicated than just
        installing the public key, we might want to move the code to a script
        within `copr-builder` and call it here or from `copr-rpmbuild`. There
        is no requirement for it to be here.
        """
        if not self.job.ssh_public_keys:
            return
        self._alloc_root_ssh_connection()
        self._set_default_expiration()
        self._deploy_user_ssh()
        self._log_user_ssh_instructions()

    def _alloc_root_ssh_connection(self):
        self.log.info("Allocating root ssh connection to builder")
        self.root_ssh = SSHConnection(
            user="root",
            host=self.host.hostname,
            port=self.host.ssh_port,
            config_file=self.opts.ssh.builder_config,
            log=self.log,
        )

    def _log_user_ssh_instructions(self):
        expiration = datetime.fromtimestamp(
            self.job.started_on + USER_SSH_DEFAULT_EXPIRATION)

        self.log.info("The owner of this build can connect using:")
        self.log.info("ssh root@%s", self.host.hostname)
        self.log.info("Unless you connect to the builder and prolong its "
                      "expiration, it will be shut-down in %s",
                      expiration.strftime(DATETIME_FORMAT))
        self.log.info("After connecting, run `copr-builder help' for "
                      "complete instructions")

    def _deploy_user_ssh(self):
        """
        Deploy user public key to the builder, so that they can connect via SSH.
        """
        pubkey = shlex.quote(self.job.ssh_public_keys)
        cmd = COMMANDS["echo_authorized_keys"].format(pubkey)
        rc, _out, _err = self.root_ssh.run_expensive(cmd)
        if rc != 0:
            self.log.error("Failed to deploy user SSH key for %s",
                           self.job.project_owner)
            return
        self.log.info("Deployed user SSH key for %s", self.job.project_owner)

    def _set_default_expiration(self):
        """
        Set the default expiration time for the builder
        """
        default = self.job.started_on + USER_SSH_DEFAULT_EXPIRATION
        cmd = COMMANDS["set_expiration"].format(shlex.quote(str(default)))
        rc, _out, _err = self.root_ssh.run_expensive(cmd)
        if rc != 0:
            # This only affects the `copr-builder show` command to print unknown
            # remaining time. It won't affect the backend in terminating the
            # buidler when it is supposed to
            self.log.error("Failed to set the default expiration time")
            return

        expiration = datetime.fromtimestamp(default)
        self.log.info("The expiration time was set to %s",
                      expiration.strftime(DATETIME_FORMAT))

    def _builder_expiration(self):
        """
        Find the user preference for the builder expiration.
        """
        rc, out, _err = self.root_ssh.run_expensive(
            COMMANDS["cat_expiration"], subprocess_timeout=60)
        if rc == 0:
            try:
                return datetime.fromtimestamp(float(out))
            except ValueError:
                pass
        self.log.error("Unable to query builder expiration file")
        return None

    def _keep_alive_for_user_ssh(self):
        """
        Wait until user releases the VM or until it expires.
        """
        if not self.job.ssh_public_keys:
            return

        # We are calculating the limits from when the job started but we may
        # want to consider starting the watch when job ends.
        default = datetime.fromtimestamp(
            self.job.started_on + USER_SSH_DEFAULT_EXPIRATION)
        maxlimit = datetime.fromtimestamp(
            self.job.started_on + USER_SSH_MAX_EXPIRATION)

        # Highlight this portion of the log because it is the only part of
        # the backend.log that is directly for the end users
        self.log.info("Keeping builder alive for user SSH")

        def _keep_alive():
            previous_expiration = default
            while True:
                if self.canceled:
                    self.log.warning("Build canceled, VM will be shut-down soon")
                    break

                expiration = self._builder_expiration() or default
                if expiration != previous_expiration:
                    self.log.info("VM expiration changed to: %s",
                                  expiration.strftime(DATETIME_FORMAT))
                    previous_expiration = expiration

                now = datetime.now()
                if now > expiration:
                    self.log.warning("VM expired, it will be shut-down soon")
                    self.log.info("The expiration was %s and it is now %s",
                                  expiration.strftime(DATETIME_FORMAT),
                                  now.strftime(DATETIME_FORMAT))
                    break
                if now > maxlimit:
                    msg = "VM exceeded max limit, it will be shut-down soon"
                    self.log.warning(msg)
                    self.log.info("The max limit was %s and it is now %s",
                                  maxlimit.strftime(DATETIME_FORMAT),
                                  now.strftime(DATETIME_FORMAT))
                    break
                time.sleep(60)

        CancellableThreadTask(
            _keep_alive,
            self._cancel_task_check_request,
            self._discard_running_worker,
            check_period=CANCEL_CHECK_PERIOD,
            log=self.log,
        ).run()
        if self.canceled:
            raise BuildCanceled

    def _cleanup(self):
        """
        Clean any temporary files after a job
        Different storage solutions (e.g. Pulp) might use backend storage as a
        temporary place for fetching builder results before uploading them
        elsewhere. There are multiple reasons for this:

        1. Builders don't have Pulp credentials, so we have to use a temporary
           place on the backend storage
        2. Some actions (`self._check_build_success`, `self._parse_results`)
           need to be run on results on backend storage

        The cleanup isn't implemented in the respective storage classes because
        we don't want e.g. `PulpStorage` and others to know anything about our
        backend storage implementation.
        """
        if self.job.storage == StorageEnum.pulp:
            path = os.path.join(self.job.chroot_dir, self.job.target_dir_name)
            for filename_entry in os.scandir(path):
                filename = filename_entry.name
                if not filename.endswith(".rpm"):
                    continue
                rpm = os.path.join(path, filename)
                self.log.info("Removing %s, it is stored in Pulp", rpm)
                os.remove(rpm)

    def build(self, attempt):
        """
        Attempt to build.
        """
        failed = True

        self._wait_for_repos()
        self._cancel_if_requested()
        self._alloc_host()
        self._alloc_ssh_connection()
        self._check_vm()
        self._fill_build_info_file()
        self._cancel_if_requested()
        self._mark_running(attempt)
        self._setup_for_user_ssh()
        self._start_remote_build()
        transfer_failure = CancellableThreadTask(
            self._transfer_log_file,
            self._check_build_interrupted,
            self._discard_running_worker,
            check_period=CANCEL_CHECK_PERIOD,
            log=self.log,
        ).run()
        if self.canceled:
            raise BuildCanceled
        if self._build_timeouted():
            raise BackendError("Build timeouted")
        if self.host.ticket.failed:
            transfer_failure = "Resalloc ticket FAILED"
        if transfer_failure:
            raise BuildRetry("SSH problems when downloading live log: {}"
                             .format(transfer_failure))

        self._keep_alive_for_user_ssh()
        self._download_results()
        self._drop_host()

        # raise error if build failed
        try:
            self._check_build_success()
            # Build _succeeded_.  Do the tasks for successful run.
            failed = False
            if self.opts.do_sign:
                self._sign_built_packages()
            self._upload_results_to_storage()
            self._do_createrepo()
            self._parse_results()
            build_details = self._get_build_details(self.job)
            self.job.update(build_details)
            self.job.validate()
            self._add_pubkey()
            self._cleanup()
        except Exception as ex:
            self.log.error("Build failed: %s", ex)
            failed = True
            raise
        finally:
            self.log.info("Finished build: id=%s failed=%s timeout=%s "
                          "destdir=%s chroot=%s ", self.job.build_id,
                          failed, self.job.timeout, self.job.destdir,
                          self.job.chroot)
            self.job.status = StatusEnum("failed" if failed else "succeeded")
            register_build_result(self.opts, failed)

    def retry_the_build(self):
        """
        Indefinitely (at most on MAX_HOST_ATTEMPTS hosts though) retry
        the build if BuildRetry is raised.
        """
        attempt = 0
        seen_hosts = set()
        while True:
            try:
                return self.build(attempt)
            except (BuildRetry, SSHConnectionError) as exc:
                seen_hosts.add(self.host.hostname)
                attempt += 1
                self.log.error("Re-try request for task on '%s': %s",
                               self.host.info, str(exc))
                self._drop_host()
                if len(seen_hosts) >= MAX_HOST_ATTEMPTS:
                    raise BackendError("Three host tried without success: {}"
                                       .format(seen_hosts))
                self.log.info("Retry #%s (on other host)", attempt)
                continue

    def handle_build(self):
        """ Do the build """
        self.sender = MessageSender(self.opts, self.name, self.log)
        self._get_build_job()
        self._setup_resultdir_and_logging()
        self._mark_starting()
        return self.retry_the_build()

    def handle_task(self):
        """ called by WorkerManager (entry point) """
        try:
            self.handle_build()
        except (BackendError, BuildCanceled, CoprBackendError) as err:
            self.log.error(str(err))
        except CoprSignError as err:
            self.log.error("Copr GPG signing problems: %s", str(err))
        except Exception:  # pylint: disable=broad-except
            self.log.exception("Unexpected exception")
        finally:
            self._drop_host()
            if self.job:
                self._mark_finished()
                self._compress_logs()
            else:
                self.log.error("No job object from Frontend")
            self.redis_set_worker_flag("status", "done")



"""
The logic behind contacting Copr Frontend from Backend.  All the requests to
the /backend/ Flask blueprint should go through this FrontendClient API.
"""

import logging

from copr_common.request import SafeRequest, RequestError
from copr_backend.exceptions import FrontendClientException

# The frontend counterpart is in `backend_general:send_frontend_version`
MIN_FE_BE_API = 7

class FrontendClient:
    """
    Object to send data back to fronted
    """

    # do we block the main daemon process?
    try_indefinitely = False

    def __init__(self, opts, logger=None, try_indefinitely=False):
        self.frontend_url = "{}/backend".format(opts.frontend_base_url)
        self.frontend_auth = opts.frontend_auth
        self.try_indefinitely = try_indefinitely

        self.msg = None
        self.logger = logger

    @property
    def log(self):
        'return configured logger object, or no-op logger'
        if not self.logger:
            self.logger = logging.getLogger(__name__)
            self.logger.addHandler(logging.NullHandler())
        return self.logger

    def get(self, url_path):
        'Issue relentless GET request to Frontend'
        return self.send(url_path, method='get')

    def post(self, url_path, data):
        'Issue relentless POST request to Frontend'
        return self.send(url_path, data=data)

    def put(self, url_path, data):
        'Issue relentless POST request to Frontend'
        return self.send(url_path, data=data, method='put')

    def send(self, url_path, method='post', data=None, authenticate=True):
        """ Repeat the request until it succeeds.  """
        while True:
            response = self._send_attempt(url_path, method, data, authenticate)
            fe_be_api_version = response.headers.get("Copr-FE-BE-API-Version", 0)
            if int(fe_be_api_version) >= MIN_FE_BE_API:
                return response

            msg = "Copr FE/BE API is too old on Frontend side, %s < %s"
            if self.try_indefinitely:
                self.logger.error(msg, fe_be_api_version, MIN_FE_BE_API)
                continue
            raise FrontendClientException(msg % (fe_be_api_version, MIN_FE_BE_API))

    def _send_attempt(self, url_path, method='post', data=None, authenticate=True):
        # """
        # Repeat the request until it succeeds, or timeout is reached.
        # """
        url = "{}/{}/".format(self.frontend_url, url_path)
        auth = ("user", self.frontend_auth) if authenticate else None
        self.log.info("Sending %s request to frontend URL - %s",
                      method.upper(), url)

        try:
            request = SafeRequest(auth=auth, log=self.log,
                                  try_indefinitely=self.try_indefinitely)
            response = request.send(url, method=method, data=data)
            return response
        except RequestError as ex:
            raise FrontendClientException(str(ex)) from ex

    def update(self, data):
        """
        Send data to be updated in the frontend
        """
        self.post("update", data)

    def starting_build(self, data):
        """
        Announce to the frontend that a build is starting.

        :return: True if the build can start or False if the build can not start (can be cancelled or deleted).
        """
        response = self.post("starting_build", data)
        if "can_start" not in response.json():
            raise FrontendClientException("Bad response from the frontend")
        return response.json()["can_start"]

    def reschedule_build(self, build_id, task_id, chroot_name):
        """
        Announce to the frontend that a build should be rescheduled (set pending state).
        """
        data = {"build_id": build_id, "task_id": task_id, "chroot": chroot_name}
        self.post("reschedule_build_chroot", data)



""" tests for BackendConfigReader class """

import shutil
import tempfile

import pytest

from copr_backend.exceptions import CoprBackendError
from copr_backend.helpers import BackendConfigReader

class TestBackendConfigReader:
    minimal_config_snippet = (
        "[backend]\n"
        "destdir=/tmp\n"
    )
    workdir = None

    def setup_method(self, method):
        _side_effects = [method]
        self.workdir = tempfile.mkdtemp(prefix="copr-test-config-")

    def teardown_method(self, method):
        _side_effects = [method]
        shutil.rmtree(self.workdir)

    def get_config_file(self, file_contents):
        """ return filename which has the configuration """
        tfile = tempfile.NamedTemporaryFile(mode="w", dir=self.workdir, delete=False)
        tfile.file.write(file_contents)
        tfile.close()
        return tfile.name

    def get_minimal_config_file(self):
        """ return filename with any (minimal) configuration """
        return self.get_config_file(self.minimal_config_snippet)

    def test_minimal_file_and_defaults(self):
        opts = BackendConfigReader(self.get_minimal_config_file()).read()
        assert opts.destdir == "/tmp"
        assert opts.builds_limits == {'arch': {}, 'tag': {}, 'owner': 20,
                                      'sandbox': 10, 'arch_per_owner': {},
                                      'userssh': 2, 'blocked_owners': []}

    def test_correct_build_limits(self):
        opts = BackendConfigReader(
            self.get_config_file(
                self.minimal_config_snippet + (
                    "builds_max_workers= 666\n"
                    "builds_max_workers_arch= x86_64 =5, aarch64= 11\n"
                    "builds_max_workers_tag = Power9=9\n"
                    "builds_max_workers_owner = 5\n"
                    "builds_max_workers_sandbox = 3\n"
                    "builds_max_workers_arch_per_owner = ppc64le=11, s390x=5\n"
                    "builds_max_userssh = 7\n"
                ))).read()
        assert opts.builds_limits == {
            'arch': {
                'x86_64': 5,
                'aarch64': 11,
            },
            'tag': {
                'Power9': 9,
            },
            'owner': 5,
            'sandbox': 3,
            'arch_per_owner': {
                'ppc64le': 11,
                's390x': 5,
            },
            'userssh': 7,
            'blocked_owners': [],
        }

    @pytest.mark.parametrize("broken_config", [
        "builds_max_workers=asdfa\n",
        "actions_max_workers=asdfa\n",
    ])
    def test_invalid_limits(self, broken_config):
        config = self.minimal_config_snippet + broken_config
        with pytest.raises(ValueError):
            BackendConfigReader(self.get_config_file(config)).read()

    @pytest.mark.parametrize("broken_config", [
        "builds_max_workers_arch=abc\n",
        "builds_max_workers_arch=abc=asdf\n",
        "builds_max_workers_arch=abc=10=\n",
        "builds_max_workers_arch=abc=1,\n",
        "builds_max_workers_tag=abc=1,\n",
    ])
    def test_invalid_build_limits(self, broken_config):
        config = self.minimal_config_snippet + broken_config
        with pytest.raises(CoprBackendError):
            BackendConfigReader(self.get_config_file(config)).read()



# pylint: disable=too-many-lines

import os
import json
import tempfile
import shutil
import time
import tarfile
import subprocess
import unittest

from unittest import mock
from unittest.mock import MagicMock

import httpretty
from munch import Munch
import pytest
from requests import RequestException

from copr_common.enums import ActionTypeEnum, BackendResultEnum, StorageEnum

from testlib.repodata import load_primary_xml
from copr_backend.actions import Action
from copr_backend.exceptions import CoprKeygenRequestError


RESULTS_ROOT_URL = "http://example.com/results"
STDOUT = "stdout"
STDERR = "stderr"


@mock.patch("copr_backend.actions.time")
class TestAction(object):

    def setup_method(self, method):
        self.tmp_dir_name = None

        self.test_time = time.time()

        self.pkgs = ["foo.src.rpm"]
        self.pkgs_stripped = ["foo"]

        self.ext_data_for_delete_build = json.dumps({
            "src_pkg_name": self.pkgs_stripped[0],
            "ownername": "foo",
            "projectname": "bar",
            "appstream": True,
            "project_dirname": "bar",
            "chroot_builddirs": {
                "fedora20": ["00001-foo"],
                "epel7": ["00001-foo"],
            },
        })

        self.opts = Munch(
            redis_db=9,
            redis_port=7777,

            destdir="/var/lib/copr/public_html/results/",
            frontend_base_url="https://example.com",
            results_baseurl=RESULTS_ROOT_URL,

            do_sign=False,

            keygen_host="example.com"
        )

        self.lockpath = tempfile.mkdtemp(prefix="copr-test-lockpath")
        self.os_env_patcher = mock.patch.dict(os.environ, {
            'COPR_TESTSUITE_NO_OUTPUT': '1',
            'COPR_TESTSUITE_LOCKPATH': self.lockpath,
            'PATH': os.environ['PATH']+':run',
        })
        self.os_env_patcher.start()

    def teardown_method(self, method):
        self.rm_tmp_dir()
        shutil.rmtree(self.lockpath)
        self.os_env_patcher.stop()

    def rm_tmp_dir(self):
        if self.tmp_dir_name:
            shutil.rmtree(self.tmp_dir_name)
            self.tmp_dir_name = None

    def make_temp_dir(self):
        root_tmp_dir = tempfile.gettempdir()
        subdir = "test_action_{}".format(time.time())
        self.tmp_dir_name = os.path.join(root_tmp_dir, subdir)

        os.mkdir(self.tmp_dir_name)

        self.test_project_dir = os.path.join(self.tmp_dir_name, "foo", "bar")
        os.makedirs(self.test_project_dir)

        # nothing should ever delete these
        os.mkdir(os.path.join(self.tmp_dir_name, "foo", "old_dir"))
        os.mkdir(os.path.join(self.tmp_dir_name, "old_dir"))

        self.test_content = "time: {}\n".format(self.test_time)

        return self.tmp_dir_name

    def unpack_resource(self, resource_name):
        if self.tmp_dir_name is None:
            self.make_temp_dir()

        src_path = os.path.join(os.path.dirname(__file__),
                                "_resources", resource_name)

        extract_to = self.test_project_dir

        with tarfile.open(src_path, "r:gz") as tfile:
            tfile.extractall(self.test_project_dir, filter="data")

        if resource_name == "testresults.tar.gz":
            # This tar.gz is inconsistently generated, but changing it would
            # only waste the git repo size.  NOTE! If you ever have to touch
            # this tarball, start distributing that tarball as separate
            # file/project and use it as Source1: in the spec file.
            testresults = os.path.join(self.test_project_dir, 'testresults')
            for subdir in os.scandir(testresults):
                shutil.move(os.path.join(testresults, subdir.name),
                            os.path.join(self.tmp_dir_name, subdir.name))
            self.test_project_dir = os.path.join(self.tmp_dir_name, '@copr', 'prunerepo')

    def test_action_run_legal_flag(self, mc_time):
        mc_time.time.return_value = self.test_time
        test_action = Action.create_from(
            opts=self.opts,
            action={
                "action_type": ActionTypeEnum("legal-flag"),
                "id": 1
            },
        )
        test_action.run()

        self.dummy = str(test_action)

    @mock.patch("copr_backend.actions.os.makedirs")
    @mock.patch("copr_backend.storage.copy_tree")
    @mock.patch("copr_backend.actions.os.path.exists")
    @mock.patch("copr_backend.storage.resign_rpms_in_dir")
    @mock.patch("copr_backend.helpers.subprocess.Popen")
    def test_action_handle_forks(self, mc_popen, mc_unsign_rpms_in_dir,
                                 mc_exists, mc_copy_tree, _mc_os_makedirs,
                                 mc_time):
        mc_popen.return_value.communicate.return_value = ("", "")
        mc_popen.return_value.returncode = 0   # Why?

        mc_time.time.return_value = self.test_time
        mc_exists = True
        test_action = Action.create_from(
            opts=self.opts,
            action={
                "action_type": ActionTypeEnum("fork"),
                "id": 1,
                "object_type": "copr",
                "data": json.dumps(
                    {
                       'builds_map':
                        {
                            'srpm-builds': {
                                '00000002': '00000009', '00000005': '00000010'},
                            'fedora-17-x86_64': {
                                '00000002-pkg1': '00000009-pkg1', '00000005-pkg2': '00000010-pkg2'},
                            'fedora-17-i386': {
                                '00000002-pkg1': '00000009-pkg1', '00000005-pkg2': '00000010-pkg2'}
                        },
                        "user": "thrnciar",
                        "copr": "source-copr",
                        "appstream": True,
                    }),
                "old_value": "thrnciar/source-copr",
                "new_value": "thrnciar/destination-copr",
            },
        )
        test_action.run()
        calls = mc_copy_tree.call_args_list
        assert len(calls) == 6
        assert calls[0][0] == (
            "/var/lib/copr/public_html/results/thrnciar/source-copr/srpm-builds/00000002",
            "/var/lib/copr/public_html/results/thrnciar/destination-copr/srpm-builds/00000009")
        assert calls[1][0] == (
            "/var/lib/copr/public_html/results/thrnciar/source-copr/srpm-builds/00000005",
            "/var/lib/copr/public_html/results/thrnciar/destination-copr/srpm-builds/00000010")
        assert calls[2][0] == (
            "/var/lib/copr/public_html/results/thrnciar/source-copr/fedora-17-x86_64/00000002-pkg1",
            "/var/lib/copr/public_html/results/thrnciar/destination-copr/fedora-17-x86_64/00000009-pkg1")
        assert calls[3][0] == (
            "/var/lib/copr/public_html/results/thrnciar/source-copr/fedora-17-x86_64/00000005-pkg2",
            "/var/lib/copr/public_html/results/thrnciar/destination-copr/fedora-17-x86_64/00000010-pkg2")
        assert calls[4][0] == (
            "/var/lib/copr/public_html/results/thrnciar/source-copr/fedora-17-i386/00000002-pkg1",
            "/var/lib/copr/public_html/results/thrnciar/destination-copr/fedora-17-i386/00000009-pkg1")
        assert calls[5][0] == (
            "/var/lib/copr/public_html/results/thrnciar/source-copr/fedora-17-i386/00000005-pkg2",
            "/var/lib/copr/public_html/results/thrnciar/destination-copr/fedora-17-i386/00000010-pkg2")

        assert len(mc_popen.call_args_list) == 2

        dirs = set()
        for call in mc_popen.call_args_list:
            args = call[0][0]
            assert args[0] == 'copr-repo'
            dirs.add(args[2])

        for chroot in ['fedora-17-i386', 'fedora-17-x86_64']:
            dir = '/var/lib/copr/public_html/results/thrnciar/destination-copr/' + chroot
            assert dir in dirs

    @unittest.skip("Fixme, test doesn't work.")
    def test_action_run_rename(self, mc_time):

        mc_time.time.return_value = self.test_time
        mc_front_cb = MagicMock()

        tmp_dir = self.make_temp_dir()
        with open(os.path.join(self.tmp_dir_name, "old_dir", "foobar.txt"), "w") as handle:
            handle.write(self.test_content)

        self.opts.destdir = tmp_dir
        test_action = Action.create_from(
            opts=self.opts,
            action={
                "action_type": ActionTypeEnum("rename"),
                "id": 1,
                "old_value": "old_dir",
                "new_value": "new_dir"
            },
            frontend_client=mc_front_cb,
        )
        test_action.run()
        result_dict = mc_front_cb.update.call_args[0][0]["actions"][0]

        assert result_dict["id"] == 1
        assert result_dict["result"] == BackendResultEnum("success")
        assert result_dict["job_ended_on"] == self.test_time

        assert not os.path.exists(os.path.join(tmp_dir, "old_dir"))
        assert os.path.exists(os.path.join(tmp_dir, "new_dir"))
        assert os.path.exists(os.path.join(tmp_dir, "new_dir", "foobar.txt"))
        with open(os.path.join(tmp_dir, "new_dir", "foobar.txt")) as handle:
            assert handle.read() == self.test_content

    @unittest.skip("Fixme, test doesn't work.")
    def test_action_run_rename_success_on_empty_src(self, mc_time):
        mc_time.time.return_value = self.test_time
        mc_front_cb = MagicMock()

        tmp_dir = self.make_temp_dir()

        self.opts.destdir = os.path.join(tmp_dir, "dir-not-exists")
        test_action = Action.create_from(
            opts=self.opts,
            action={
                "action_type": ActionTypeEnum("rename"),
                "id": 1,
                "old_value": "old_dir",
                "new_value": "new_dir"
            },
            frontend_client=mc_front_cb,
        )
        test_action.run()
        result_dict = mc_front_cb.update.call_args[0][0]["actions"][0]

        assert result_dict["id"] == 1
        assert result_dict["result"] == BackendResultEnum("success")
        assert result_dict["job_ended_on"] == self.test_time

        assert os.path.exists(os.path.join(tmp_dir, "old_dir"))

    @unittest.skip("Fixme, test doesn't work.")
    def test_action_run_rename_failure_on_destination_exists(self, mc_time):
        mc_time.time.return_value = self.test_time
        mc_front_cb = MagicMock()

        tmp_dir = self.make_temp_dir()
        with open(os.path.join(self.tmp_dir_name, "old_dir", "foobar.txt"), "w") as handle:
            handle.write(self.test_content)
        os.mkdir(os.path.join(tmp_dir, "new_dir"))

        self.opts.destdir = tmp_dir
        test_action = Action.create_from(
            opts=self.opts,
            action={
                "action_type": ActionTypeEnum("rename"),
                "id": 1,
                "old_value": "old_dir",
                "new_value": "new_dir"
            },
            frontend_client=mc_front_cb,
        )
        test_action.run()
        result_dict = mc_front_cb.update.call_args[0][0]["actions"][0]

        assert result_dict["id"] == 1
        assert result_dict["result"] == BackendResultEnum("failure")
        assert result_dict["message"] == "Destination directory already exist."
        assert result_dict["job_ended_on"] == self.test_time

        assert os.path.exists(os.path.join(tmp_dir, "old_dir"))
        assert os.path.exists(os.path.join(tmp_dir, "new_dir"))
        assert not os.path.exists(os.path.join(tmp_dir, "new_dir", "foobar.txt"))

    def test_action_run_delete_copr(self, mc_time):
        tmp_dir = self.make_temp_dir()
        self.opts.destdir = tmp_dir

        test_action = Action.create_from(
            opts=self.opts,
            action={
                "action_type": ActionTypeEnum("delete"),
                "object_type": "copr",
                "id": 6,
                # baz doesn't exist
                "data": json.dumps({
                    "ownername": "foo",
                    "project_dirnames": ["bar", "baz"],
                    "appstream": True,
                }),
            },
        )

        assert os.path.exists(os.path.join(tmp_dir, "foo", "bar"))
        assert not os.path.exists(os.path.join(tmp_dir, "foo", "baz"))
        assert test_action.run() == BackendResultEnum("success")
        assert os.path.exists(os.path.join(tmp_dir, "old_dir"))
        assert not os.path.exists(os.path.join(tmp_dir, "foo", "bar"))

    @unittest.skip("Fixme, test doesn't work.")
    def test_action_run_delete_copr_remove_folders(self, mc_time):
        mc_time.time.return_value = self.test_time
        mc_front_cb = MagicMock()

        tmp_dir = self.make_temp_dir()
        self.opts.destdir=tmp_dir
        test_action = Action.create_from(
            opts=self.opts,
            action={
                "action_type": ActionTypeEnum("delete"),
                "object_type": "copr",
                "id": 6,
                "old_value": "old_dir",
            },
            frontend_client=mc_front_cb
        )
        test_action.run()

        result_dict = mc_front_cb.update.call_args[0][0]["actions"][0]
        assert result_dict["id"] == 6
        assert result_dict["result"] == BackendResultEnum("success")
        assert result_dict["job_ended_on"] == self.test_time

        assert not os.path.exists(os.path.join(tmp_dir, "old_dir"))

    @mock.patch("copr_backend.actions.call_copr_repo")
    def test_delete_no_chroot_dirs(self, mc_call, mc_time):
        mc_time.time.return_value = self.test_time
        mc_front_cb = MagicMock()

        tmp_dir = self.make_temp_dir()
        self.opts.destdir = tmp_dir
        test_action = Action.create_from(
            opts=self.opts,
            action={
                "id": 1,
                "object_id": 1,
                "action_type": ActionTypeEnum("delete"),
                "object_type": "build",
                "data": self.ext_data_for_delete_build,
            },

        )
        result = test_action.run()
        assert len(mc_call.call_args_list) == 0
        assert result == BackendResultEnum("failure")

    def test_delete_build_succeeded(self, mc_time):
        mc_time.time.return_value = self.test_time
        mc_front_cb = MagicMock()

        tmp_dir = self.make_temp_dir()

        proj_dir = os.path.join(tmp_dir, "foo", "bar")
        chroot_1_dir = os.path.join(proj_dir, "fedora20")
        chroot_2_dir = os.path.join(proj_dir, "epel7")
        foo_pkg_dir = os.path.join(chroot_1_dir, "00001-foo")

        os.makedirs(chroot_2_dir)
        os.makedirs(foo_pkg_dir)

        with open(os.path.join(foo_pkg_dir, "foo.src.rpm"), "w") as fh:
            fh.write("foo\n")

        log_path = os.path.join(chroot_1_dir, "build-{:08d}.log".format(42))
        with open(log_path, "w") as fh:
            fh.write(self.test_content)

        self.opts.destdir = tmp_dir
        test_action = Action.create_from(
            opts=self.opts,
            action={
                "action_type": ActionTypeEnum("delete"),
                "object_type": "build",
                "id": 7,
                "old_value": "old_dir",
                "data": self.ext_data_for_delete_build,
                "object_id": 42
            },
        )

        assert os.path.exists(foo_pkg_dir)
        assert test_action.run() == BackendResultEnum("success")
        assert not os.path.exists(foo_pkg_dir)
        assert not os.path.exists(log_path)
        assert os.path.exists(chroot_1_dir)
        assert os.path.exists(chroot_2_dir)

    @pytest.mark.parametrize('devel', [False, True])
    def test_delete_build_acr_reflected(self, mc_time, devel):
        """
        When build is deleted, we want to remove it from both devel and normal
        (production) repodata
        """
        self.unpack_resource("testresults.tar.gz")

        chroot = os.path.join(self.test_project_dir, 'fedora-23-x86_64')
        repodata = os.path.join(chroot, 'repodata')
        devel_dir = os.path.join(chroot, 'devel')
        repodata_devel = os.path.join(devel_dir, 'repodata')
        builddir = '00000049-example'
        assert os.path.exists(chroot)
        assert not os.path.exists(devel_dir)

        # create repodata under 'devel'
        assert subprocess.call(['copr-repo', chroot, '--devel']) == 0

        old_primary = load_primary_xml(repodata)
        old_primary_devel = load_primary_xml(repodata_devel)
        assert len(old_primary['names']) == 3 # noarch vs. src
        assert len(old_primary['hrefs']) == 5

        for package in old_primary_devel['packages']:
            assert old_primary_devel['packages'][package]['xml:base'] \
                   == 'https://example.com/results/@copr/prunerepo/fedora-23-x86_64'
            # clear it
            old_primary_devel['packages'][package]['xml:base'] = ''

        assert old_primary == old_primary_devel

        self.opts.destdir = self.tmp_dir_name

        test_action = Action.create_from(
            opts=self.opts,
            action={
                "action_type": ActionTypeEnum("delete"),
                "object_type": "build",
                "id": 7,
                "object_id": 49,
                "old_value": "old_dir",
                "data": json.dumps({
                    "ownername": "@copr",
                    "projectname": "prunerepo",
                    "project_dirname": "prunerepo",
                    "appstream": True,
                    "chroot_builddirs": {
                        "fedora-23-x86_64": [builddir],
                    },
                    "devel": devel,
                }),
            },
        )

        assert test_action.run() == BackendResultEnum("success")

        new_primary = load_primary_xml(repodata)
        new_primary_devel = load_primary_xml(repodata_devel)

        if devel:
            assert new_primary_devel['names'] == set(['prunerepo'])
            assert len(new_primary['names']) == 3
        else:
            assert new_primary['names'] == set(['prunerepo'])
            assert len(new_primary_devel['names']) == 3

    @mock.patch("copr_backend.storage.call_copr_repo")
    def test_delete_build_succeeded_createrepo_error(self,
                                                     mc_call_repo, mc_time):
        mc_time.time.return_value = self.test_time
        mc_front_cb = MagicMock()

        mc_call_repo.return_value = 0
        tmp_dir = self.make_temp_dir()

        chroot_1_dir = os.path.join(tmp_dir, "foo", "bar", "fedora20")
        foo_pkg_dir = os.path.join(chroot_1_dir, "foo")
        os.makedirs(foo_pkg_dir)
        chroot_2_dir = os.path.join(tmp_dir, "foo", "bar", "epel7")
        os.mkdir(chroot_2_dir)

        with open(os.path.join(foo_pkg_dir, "foo.src.rpm"), "w") as fh:
            fh.write("foo\n")

        self.opts.destdir = tmp_dir
        test_action = Action.create_from(
            opts=self.opts,
            action={
                "action_type": ActionTypeEnum("delete"),
                "object_type": "build",
                "id": 7,
                "old_value": "old_dir",
                "data": self.ext_data_for_delete_build,
                "object_id": 42,
            },
        )
        # just fail
        assert test_action.run() == BackendResultEnum("failure")

    def test_delete_two_chroots(self, mc_time):
        """
        Regression test, https://bugzilla.redhat.com/show_bug.cgi?id=1171796
        """
        self.unpack_resource("1171796.tar.gz")

        chroot_20_path = os.path.join(self.tmp_dir_name, "foo", "bar", "fedora-20-x86_64")
        chroot_21_path = os.path.join(self.tmp_dir_name, "foo", "bar", "fedora-21-x86_64")

        assert os.path.exists(os.path.join(chroot_20_path, "build-15.log"))
        assert os.path.exists(os.path.join(chroot_21_path, "build-15.log"))

        for path in [chroot_20_path, chroot_21_path]:
            # the 1171796 tarball above contains too old directory structure,
            # for some time we prefix the ID with zeroes.  So create another
            # (newer variant of) log file and check that both logs would be
            # potentially removed.
            shutil.copy(os.path.join(path, "build-15.log"),
                        os.path.join(path, "build-00000015.log"))

        assert os.path.exists(os.path.join(chroot_20_path, "build-15.rsync.log"))
        assert os.path.exists(os.path.join(chroot_21_path, "build-15.rsync.log"))

        assert os.path.isdir(os.path.join(chroot_20_path, "rubygem-log4r-1.1.10-2.fc21"))
        assert os.path.isdir(os.path.join(chroot_21_path, "rubygem-log4r-1.1.10-2.fc21"))

        mc_time.time.return_value = self.test_time

        self.opts.destdir = self.tmp_dir_name
        test_action = Action.create_from(
            opts=self.opts,
            action={
                "action_type": ActionTypeEnum("delete"),
                "object_type": "build",
                "object_id": 15,
                "id": 15,
                "old_value": "old_dir",
                "data": json.dumps({
                    "ownername": "foo",
                    "projectname": "bar",
                    "appstream": True,
                    "project_dirname": "bar",
                    "chroot_builddirs": {
                        "fedora-20-x86_64": ["rubygem-log4r-1.1.10-2.fc21"],
                        "fedora-21-x86_64": ["rubygem-log4r-1.1.10-2.fc21"],
                    }
                }),
            },
        )
        assert test_action.run() == BackendResultEnum("success")

        assert not os.path.exists(os.path.join(chroot_20_path, "build-00000015.log"))
        assert not os.path.exists(os.path.join(chroot_21_path, "build-00000015.log"))
        assert not os.path.exists(os.path.join(chroot_20_path, "build-15.log"))
        assert not os.path.exists(os.path.join(chroot_21_path, "build-15.log"))

        assert not os.path.exists(os.path.join(chroot_20_path, "build-15.rsync.log"))
        assert not os.path.exists(os.path.join(chroot_21_path, "build-15.rsync.log"))

        assert not os.path.isdir(os.path.join(chroot_20_path, "rubygem-log4r-1.1.10-2.fc21"))
        assert not os.path.isdir(os.path.join(chroot_21_path, "rubygem-log4r-1.1.10-2.fc21"))

        assert os.path.exists(chroot_20_path)
        assert os.path.exists(chroot_21_path)

    def test_delete_two_chroots_two_remain(self, mc_time):
        """
        Regression test, https://bugzilla.redhat.com/show_bug.cgi?id=1171796
        extended: we also put two more chroots, which should be unaffected
        """
        self.unpack_resource("1171796_doubled.tar.gz")

        subdir = os.path.join(self.tmp_dir_name, "foo", "bar")

        chroot_20_path = os.path.join(subdir, "fedora-20-x86_64")
        chroot_21_path = os.path.join(subdir, "fedora-21-x86_64")
        chroot_20_i386_path = os.path.join(subdir, "fedora-20-i386")
        chroot_21_i386_path = os.path.join(subdir, "fedora-21-i386")

        assert os.path.exists(os.path.join(chroot_20_path, "build-15.log"))
        assert os.path.exists(os.path.join(chroot_21_path, "build-15.log"))
        assert os.path.exists(os.path.join(chroot_20_i386_path, "build-15.log"))
        assert os.path.exists(os.path.join(chroot_21_i386_path, "build-15.log"))

        assert os.path.exists(os.path.join(chroot_20_path, "build-15.rsync.log"))
        assert os.path.exists(os.path.join(chroot_21_path, "build-15.rsync.log"))
        assert os.path.exists(os.path.join(chroot_20_i386_path, "build-15.rsync.log"))
        assert os.path.exists(os.path.join(chroot_21_i386_path, "build-15.rsync.log"))

        assert os.path.isdir(os.path.join(chroot_20_path, "rubygem-log4r-1.1.10-2.fc21"))
        assert os.path.isdir(os.path.join(chroot_21_path, "rubygem-log4r-1.1.10-2.fc21"))
        assert os.path.isdir(os.path.join(chroot_20_i386_path, "rubygem-log4r-1.1.10-2.fc21"))
        assert os.path.isdir(os.path.join(chroot_21_i386_path, "rubygem-log4r-1.1.10-2.fc21"))

        mc_time.time.return_value = self.test_time

        self.opts.destdir = self.tmp_dir_name
        test_action = Action.create_from(
            opts=self.opts,
            action={
                "action_type": ActionTypeEnum("delete"),
                "object_type": "build",
                "object_id": 15,
                "id": 15,
                "old_value": "old_dir",
                "data": json.dumps({
                    "ownername": "foo",
                    "projectname": "bar",
                    "appstream": True,
                    "chroots": ["fedora-20-x86_64", "fedora-21-x86_64"],
                    "project_dirname": "bar",
                    "chroot_builddirs": {
                        "fedora-20-x86_64": ["rubygem-log4r-1.1.10-2.fc21"],
                        "fedora-21-x86_64": ["rubygem-log4r-1.1.10-2.fc21"],
                    }
                }),
            },
        )

        assert test_action.run() == BackendResultEnum("success")

        assert not os.path.exists(os.path.join(chroot_20_path, "build-15.log"))
        assert not os.path.exists(os.path.join(chroot_21_path, "build-15.log"))
        assert os.path.exists(os.path.join(chroot_20_i386_path, "build-15.log"))
        assert os.path.exists(os.path.join(chroot_21_i386_path, "build-15.log"))

        assert not os.path.exists(os.path.join(chroot_20_path, "build-15.rsync.log"))
        assert not os.path.exists(os.path.join(chroot_21_path, "build-15.rsync.log"))
        assert os.path.exists(os.path.join(chroot_20_i386_path, "build-15.rsync.log"))
        assert os.path.exists(os.path.join(chroot_21_i386_path, "build-15.rsync.log"))

        assert not os.path.isdir(os.path.join(chroot_20_path, "rubygem-log4r-1.1.10-2.fc21"))
        assert not os.path.isdir(os.path.join(chroot_21_path, "rubygem-log4r-1.1.10-2.fc21"))
        assert os.path.isdir(os.path.join(chroot_20_i386_path, "rubygem-log4r-1.1.10-2.fc21"))
        assert os.path.isdir(os.path.join(chroot_21_i386_path, "rubygem-log4r-1.1.10-2.fc21"))

    def test_delete_build_with_bad_pkg_name(self, mc_time):
        """
        Originally written for regression:
        https://bugzilla.redhat.com/show_bug.cgi?id=1203753
        But we nowadays donjt send src_pkg_name at all.
        """
        mc_time.time.return_value = self.test_time

        self.unpack_resource("1171796_doubled.tar.gz")

        chroot_20_path = os.path.join(self.tmp_dir_name, "foo", "bar", "fedora-20-x86_64")
        chroot_21_path = os.path.join(self.tmp_dir_name, "foo", "bar", "fedora-21-x86_64")

        self.opts.destdir = self.tmp_dir_name
        test_action = Action.create_from(
            opts=self.opts,
            action={
                "action_type": ActionTypeEnum("delete"),
                "object_type": "build",
                "object_id": 15,
                "id": 15,
                "old_value": "old_dir",
                "data": json.dumps({
                    "src_pkg_name": "",
                    "ownername": "foo",
                    "projectname": "bar",
                    "project_dirname": "bar",
                    "chroots": ["fedora-20-x86_64", "fedora-21-x86_64"]
                }),
            },
        )

        assert os.path.exists(chroot_20_path)
        assert os.path.exists(chroot_21_path)
        result = test_action.run()
        assert result == BackendResultEnum("failure")

        # shouldn't touch chroot dirs
        assert os.path.exists(chroot_20_path)
        assert os.path.exists(chroot_21_path)

    def test_delete_multiple_builds_succeeded(self, mc_time):
        mc_time.time.return_value = self.test_time

        tmp_dir = self.make_temp_dir()

        chroot_dir = os.path.join(tmp_dir, "foo", "bar", "fedora-20")
        os.makedirs(chroot_dir)
        pkg_build_1_dir = os.path.join(chroot_dir, "01-foo")
        pkg_build_2_dir = os.path.join(chroot_dir, "02-foo")
        pkg_build_3_dir = os.path.join(chroot_dir, "03-foo")
        os.mkdir(pkg_build_1_dir)
        os.mkdir(pkg_build_2_dir)
        os.mkdir(pkg_build_3_dir)

        ext_data = json.dumps({
            "ownername": "foo",
            "projectname": "bar",
            "appstream": True,
            "devel": False,
            "project_dirnames": {
                'bar': {
                    "fedora-20": ["01-foo", "02-foo"],
                }
            },
            "build_ids": ['01', '02'],
        })

        self.opts.destdir = tmp_dir
        test_action = Action.create_from(
            opts=self.opts,
            action={
                "action_type": ActionTypeEnum("delete"),
                "object_type": "builds",
                "id": 7,
                "data": ext_data,
            },
        )

        assert os.path.exists(pkg_build_1_dir)
        assert os.path.exists(pkg_build_2_dir)
        test_action.run()
        assert not os.path.exists(pkg_build_1_dir)
        assert not os.path.exists(pkg_build_2_dir)
        assert os.path.exists(chroot_dir)
        assert os.path.exists(pkg_build_3_dir)

    # We want to test that ACR flag doesn't make any difference here, explicit
    # createrepo always works with non-devel directory.
    @pytest.mark.parametrize('devel', [False, True])
    @mock.patch("copr_backend.helpers.subprocess.Popen")
    def test_handle_createrepo_ok(self, mc_sp_popen, mc_time, devel):
        mc_sp_popen.return_value.communicate.return_value = ("", "")
        mc_sp_popen.return_value.returncode = 0

        tmp_dir = self.make_temp_dir()

        action_data = json.dumps({
            "chroots": ["epel-6-i386", "fedora-20-x86_64"],
            "ownername": "foo",
            "projectname": "bar",
            "appstream": True,
            "project_dirnames": ["bar"],
            "devel": devel,
            "storage": StorageEnum.backend,
            "reason": None,
        })
        self.opts.destdir = tmp_dir

        test_action = Action.create_from(
            opts=self.opts,
            action={
                "action_type": ActionTypeEnum("createrepo"),
                "data": action_data,
                "id": 8
            },
        )
        assert test_action.run() == BackendResultEnum("success")

        for chroot in ['fedora-20-x86_64', 'epel-6-i386']:
            cmd = ["copr-repo", "--batched",
                   os.path.join(self.test_project_dir, chroot)]
            if devel:
                cmd.append("--devel")
            exp_call = mock.call(cmd, stdout=-1, stderr=-1, shell=False, encoding='utf-8')
            assert exp_call in mc_sp_popen.call_args_list

        assert len(mc_sp_popen.call_args_list) == 2

    @mock.patch("copr_backend.storage.call_copr_repo")
    def test_handle_createrepo_failure_1(self, mc_call, mc_time):
        tmp_dir = self.make_temp_dir()
        mc_call.return_value = 0 # failure

        action_data = json.dumps({
            "chroots": ["epel-6-i386", "fedora-20-x86_64"],
            "ownername": "foo",
            "projectname": "bar",
            "appstream": True,
            "project_dirnames": ["bar"],
            "devel": False,
            "storage": StorageEnum.backend,
            "reason": None,
        })
        self.opts.destdir = tmp_dir
        test_action = Action.create_from(
            opts=self.opts,
            action={
                "action_type": ActionTypeEnum("createrepo"),
                "data": action_data,
                "id": 9
            },
        )
        assert test_action.run() == BackendResultEnum("failure")

    @unittest.skip("Fixme, test doesn't work.")
    @mock.patch("copr_backend.actions.create_user_keys")
    def test_handle_generate_gpg_key(self, mc_cuk, mc_time):
        uname = "foo"
        pname = "bar"
        action_data = json.dumps({
            "username": uname,
            "projectname": pname,
        })

        expected_call = mock.call(uname, pname, self.opts)

        mc_front_cb = MagicMock()
        test_action = Action.create_from(
            opts=self.opts,
            action={
                "action_type": ActionTypeEnum("gen_gpg_key"),
                "data": action_data,
                "id": 11
            },
            frontend_client=mc_front_cb
        )

        test_action.run()

        result_dict = mc_front_cb.update.call_args[0][0]["actions"][0]
        assert result_dict["id"] == 11
        assert result_dict["result"] == BackendResultEnum("success")

        assert mc_cuk.call_args == expected_call

        # handle exception
        mc_cuk.side_effect = CoprKeygenRequestError("foo")
        mc_front_cb.reset_mock()
        test_action.run()

        result_dict = mc_front_cb.update.call_args[0][0]["actions"][0]

        assert result_dict["id"] == 11
        assert result_dict["result"] == BackendResultEnum("failure")

        # test, that key creation is skipped when signing is disabled
        self.opts.do_sign = False
        mc_front_cb.reset_mock()
        mc_cuk.reset_mock()
        test_action.run()

        assert not mc_cuk.called
        result_dict = mc_front_cb.update.call_args[0][0]["actions"][0]
        assert result_dict["id"] == 11
        assert result_dict["result"] == BackendResultEnum("success")

    @unittest.skip("Fixme, test doesn't work.")
    def test_request_exception_is_taken_care_of_when_posting_to_frontend(self, mc_time):
        mc_time.time.return_value = self.test_time
        mc_frontend_client = MagicMock()
        mc_frontend_client.update = MagicMock(side_effect=RequestException)

        tmp_dir = self.make_temp_dir()
        self.opts.destdir = os.path.join(tmp_dir, "dir-not-exists")

        test_action = Action.create_from(
            opts=self.opts,
            action={
                "action_type": ActionTypeEnum("delete"),
                "object_type": "copr",
                "id": 6,
                "old_value": "old_dir",
            },
            frontend_client=mc_frontend_client,
        )

        try:
            test_action.run()
        except Exception as e:
            assert False

    def test_delete_chroot(self, mc_time):
        mc_time.time.return_value = self.test_time
        mc_front_cb = MagicMock()

        tmp_dir = self.make_temp_dir()

        proj_dir = os.path.join(tmp_dir, "foo", "bar")
        chroot_1_dir = os.path.join(proj_dir, "fedora20")
        chroot_2_dir = os.path.join(proj_dir, "epel7")
        foo_pkg_dir = os.path.join(chroot_1_dir, "00001-foo")

        os.makedirs(chroot_2_dir)
        os.makedirs(foo_pkg_dir)

        with open(os.path.join(foo_pkg_dir, "foo.src.rpm"), "w") as fh:
            fh.write("foo\n")

        log_path = os.path.join(chroot_1_dir, "build-{:08d}.log".format(42))
        with open(log_path, "w") as fh:
            fh.write(self.test_content)
        self.opts.destdir = tmp_dir

        data = json.dumps({
            "ownername": "foo",
            "projectname": "bar",
            "chrootname": "fedora20",
        })
        test_action = Action.create_from(
            opts=self.opts,
            action={
                "action_type": ActionTypeEnum("delete"),
                "object_type": "chroot",
                "id": None,
                "old_value": None,
                "data": data,
                "object_id": None,
            },
        )

        assert os.path.exists(foo_pkg_dir)
        assert os.path.exists(chroot_1_dir)
        assert test_action.run() == BackendResultEnum("success")
        assert not os.path.exists(foo_pkg_dir)
        assert not os.path.exists(chroot_1_dir)

        # The action shouldn't fail even when the directory doesn't exist anymore
        assert test_action.run() == BackendResultEnum("success")

    @httpretty.activate()
    def test_comps_create(self, mc_time):
        _ = mc_time
        text = "some xml body\n"
        httpretty.register_uri(
            httpretty.GET,
            "https://example.com/comps/path/",
            body=text,
        )

        self.opts.destdir = self.make_temp_dir()
        action_dict = {
            "action_type": 4,
            "created_on": 1622632008,
            "data": json.dumps({
                "ownername": "praiskup",
                "projectname": "ping",
                "chroot": "fedora-rawhide-x86_64",
                "comps_present": True,
                "url_path": "/comps/path/",
            }),
            "id": 672272,
            "object_type": "copr_chroot",
            "object_id": None,
            "old_value": None,
            "priority": 0,
        }
        test_action = Action.create_from(
            opts=self.opts,
            action=action_dict,
        )
        assert test_action.run() == BackendResultEnum("success")

        file = os.path.join(self.opts.destdir,
                            "praiskup/ping/fedora-rawhide-x86_64",
                            "comps.xml")
        with open(file, "r") as fd:
            lines = fd.readlines()
            assert lines == [text]

    @mock.patch("copr_backend.actions.shutil.rmtree")
    def test_remove_dirs(self, mock_rmtree, mc_time):
        _unused = mc_time
        test_action = Action.create_from(
            opts=self.opts,
            action={
                "action_type": ActionTypeEnum("remove_dirs"),
                "data": json.dumps([
                    "@python/python3.8:pr:11",
                    "jdoe/some:pr:123",
                ]),
            },
        )
        assert test_action.run() == BackendResultEnum("success")
        assert mock_rmtree.call_args_list == [
            mock.call('/var/lib/copr/public_html/results/@python/python3.8:pr:11'),
            mock.call('/var/lib/copr/public_html/results/jdoe/some:pr:123'),
        ]



# coding: utf-8
import json
import subprocess
from subprocess import CalledProcessError

ansible_playbook_bin = "ansible-playbook"


def ans_extra_vars_encode(extra_vars, name):
    """ transform dict into --extra-vars="json string" """
    if not extra_vars:
        return ""
    return "--extra-vars='{{\"{0}\": {1}}}'".format(name, json.dumps(extra_vars))


def run_ansible_playbook_cli(args, comment, log, timeout=None):
    """
    :param args:
    :param comment:
    :param timeout:
    :type log: logging.Logger
    :return: ansible output
    """
    command = "{} -v {}".format(ansible_playbook_bin, args)

    if comment is None:
        comment = "running playbook"
    if timeout:
        command = "timeout {} {}".format(timeout, command)

    try:
        log.info("%s: begin: %s", comment, command)
        result = subprocess.check_output(command, shell=True, stderr=subprocess.STDOUT, encoding="utf-8")
    except CalledProcessError as e:
        log.info("CalledProcessError: %s", e.output)
        raise

    log.debug(comment + ": end")
    return result



class CoprSignError(Exception):
    """
    Related to invocation of /bin/sign

    has additional  fields:
    :ivar cmd: command which was executed
    :ivar stdout: message content
    :ivar stderr: error message
    """

    def __init__(self, msg, cmd=None, stdout=None, stderr=None,
                 return_code=None):

        super(CoprSignError, self).__init__(msg)
        self.cmd = cmd
        self.return_code = return_code
        self.stdout = stdout
        self.stderr = stderr

    def __str__(self):
        out = super(CoprSignError, self).__str__()
        if self.cmd:
            out += ("\n"
                    "return code {} after invocation of: {} \n"
                    "stderr: {}\n"
                    "stdout: {}\n").format(
                        self.return_code, self.cmd, self.stdout, self.stderr)
        return out


class CoprSignNoKeyError(CoprSignError):
    pass


class CoprKeygenRequestError(Exception):
    """
    Errors during request to copr-keygen service

    has additional  fields:
    :ivar request: tuple of parameters for request.request
    :ivar response: requests.Response
    """

    def __init__(self, msg, request=None, response=None):
        super(CoprKeygenRequestError, self).__init__(msg)
        self.request = request
        self.response = response

    def __str__(self):
        out = super(CoprKeygenRequestError, self).__str__()
        out += "\nrequest to copr-keygen: {}\n".format(self.request)
        if self.response:
            out += "status code: {}\n" "response content: {}\n" \
                .format(self.response.status_code, self.response.content)
        return out


class CoprBackendError(Exception):
    def __init__(self, msg):
        super(CoprBackendError, self).__init__()
        self.msg = msg

    def __str__(self):
        return self.msg


class CoprJobGrabError(CoprBackendError):
    pass


class CoprWorkerError(CoprBackendError):
    pass


class CoprSpawnFailError(CoprBackendError):
    pass

class CoprBackendSrpmError(CoprBackendError):
    pass

class VmError(CoprBackendError):
    """
    Error related to VM manage
    """
    pass


class VmDescriptorNotFound(VmError):
    pass


class NoVmAvailable(VmError):
    pass

class VmSpawnLimitReached(VmError):
    """
    Couldn't spawn more VM due to the given limits
    """
    pass


class CmdError(CoprBackendError):
    def __init__(self, msg, cmd, exit_code=None, stdout=None, stderr=None):
        super(CmdError, self).__init__(msg)

        self.cmd = cmd
        self.exit_code = exit_code
        self.stdout = stdout
        self.stderr = stderr

    def __str__(self):
        out = super(CmdError, self).__str__()
        if self.cmd:
            out += ("\n"
                    "return code {} after invocation of: {} \n"
                    "stderr: {}\n"
                    "stdout: {}\n").format(
                        self.exit_code, self.cmd, self.stdout, self.stderr)
        return out


class CreateRepoError(CmdError):
    pass


class CoprRepoCmdError(CmdError):
    pass


class DispatchBuildError(CoprBackendError):
    pass


class FrontendClientException(Exception):
    pass



import os
from logging import Formatter
from typing import Dict, List

mockchain = "/usr/bin/mockchain"
# rsync path
rsync = "/usr/bin/rsync"

DEF_BUILD_TIMEOUT = 3600 * 6
DEF_REPOS = []
DEF_CHROOT = None
DEF_BUILD_USER = "mockbuilder"
DEF_DESTDIR = os.getcwd()
DEF_MACROS = {}
DEF_BUILDROOT_PKGS = ""


DEF_CONSECUTIVE_FAILURE_THRESHOLD = 10
CONSECUTIVE_FAILURE_REDIS_KEY = "copr:sys:consecutive_build_fails"
PULP_REDIRECT_FILE = "/var/lib/copr/public_html/pulp-redirect.txt"


class BuildStatus(object):
    FAILURE = 0
    SUCCEEDED = 1
    RUNNING = 3
    PENDING = 4
    SKIPPED = 5
    STARTING = 6

    @classmethod
    def string(cls, number):
        """ convert number to string """
        for key, val in cls.__dict__.items():
            if isinstance(val, int) and number == val:
                return key
        raise AttributeError("no such status id: {0} ".format(number))


LOG_REDIS_FIFO = "copr:backend:log:fifo::"

default_log_format = Formatter(
    '[%(asctime)s][%(levelname)6s][PID:%(process)d][%(name)10s][%(filename)s:%(funcName)s:%(lineno)d] %(message)s')
build_log_format = Formatter(
    '[%(asctime)s][%(levelname)6s][PID:%(process)d] %(message)s')
script_log_format = Formatter(
    "[%(asctime)s][%(thread)s][%(levelname)6s]: %(message)s")

# dict of OS versions that are too old or the use sqlite repodata so they need --database
# option with createrepo. This dict is in format "os_family": <list of versions>
# e.g. "rhel": ["7", "6", "5"]
CHROOTS_USING_SQLITE_REPODATA: Dict[str, List[str]] = {}



"""
Test the SSHConnection class
"""

from copr_backend.sshcmd import SSHConnection

def test_ipv4_ipv6_rsync():
    connection = SSHConnection(
        "test", "2620:52:3:1:dead:beef:cafe:c149", config_file="something",
    )
    # pylint: disable=protected-access
    assert connection._full_source_path("/xyz") == "test@[2620:52:3:1:dead:beef:cafe:c149]:/xyz"
    connection = SSHConnection(
        "test", "192.168.0.1", config_file="something",
    )
    assert connection._full_source_path("/xyz") == "test@192.168.0.1:/xyz"



import json
import os

from copr_common.redis_helpers import get_redis_connection

# todo: add logging here
# from copr_backend.helpers import BackendConfigReader, get_redis_logger
# opts = BackendConfigReader().read()
# log = get_redis_logger(opts, "createrepo", "actions")

# Some reasonable limit here for exceptional (probably buggy) situations.
# This is here mostly to not overflow the execve() stack limits.
MAX_IN_BATCH = 100


class BatchedCreaterepo:
    """
    Group a "group-able" set of pending createrepo tasks, and execute
    the createrepo_c binary only once for the batch.  As a result, some
    `copr-repo` processes do slightly more work (negligible difference compared
    to overall createrepo_c cost) but some do nothing.

    Note that this is wrapped into separate class mostly to make the unittesting
    easier.

    The process goes like this:

    1. BatchedCreaterepo() is instantiated by caller.
    2. Before caller acquires createrepo lock, caller notifies other processes
       by make_request().
    3. Caller acquires createrepo lock.
    4. Caller assures that no other process already did it's task, by calling
       check_processed() method (if done, caller _ends_).  Others are now
       waiting for lock so they can not process our task in the meantime.
    5. Caller get's "unified" createrepo options that are needed by the other
       queued processes by calling options() method.  These options are then
       merged with options needed by caller's task, and createrepo_c is
       executed.  Now we are saving the resources.
    6. The commit() method is called (under lock) to notify others that they
       don't have to duplicate the efforts and waste resources.
    """
    # pylint: disable=too-many-instance-attributes
    # pylint: disable=too-many-arguments

    def __init__(self, dirname, full, add, delete, rpms_to_remove, log,
                 devel=False,
                 appstream=True,
                 backend_opts=None,
                 noop=False):
        self.noop = noop
        self.log = log
        self.dirname = dirname
        self.devel = devel
        self.appstream = appstream
        self.rpms_to_remove = rpms_to_remove

        if not backend_opts:
            self.log.error("can't get access to redis, batch disabled")
            self.noop = True
            return

        self._pid = os.getpid()
        self._json_redis_task = json.dumps({
            "appstream": appstream,
            "devel": devel,
            "add": add,
            "delete": delete,
            "full": full,
            "rpms_to_remove": rpms_to_remove,
        })

        self.notify_keys = []
        self.redis = get_redis_connection(backend_opts)

    @property
    def key(self):
        """ Our instance ID (key in Redis DB) """
        return "createrepo_batched::{}::{}".format(
            self.dirname, self._pid)

    @property
    def key_pattern(self):
        """ Redis key pattern for potential tasks we can batch-process """
        return "createrepo_batched::{}::*".format(self.dirname)

    def make_request(self):
        """ Request the task into Redis DB.  Run _before_ lock! """
        if self.noop:
            return None
        self.redis.hset(self.key, "task", self._json_redis_task)
        return self.key

    def check_processed(self, delete_if_not=True):
        """
        Drop our entry from Redis DB (if any), and return True if the task is
        already processed.  When 'delete_if_not=True, we delete the self.key
        from Redis even if the task is not yet processed (meaning that caller
        plans to finish the task right away).
        """
        if self.noop:
            return False

        status = self.redis.hget(self.key, "status") == "success"
        self.log.debug("Has already a status? %s", status)

        try:
            if not status:
                # not yet processed
                return False
        finally:
            # This is atomic operation, other processes may not re-start doing this
            # task again.  https://github.com/redis/redis/issues/9531
            #
            # The 'delete_if_not = True' is not a TOCTOU case.  Our caller holds
            # the lock, hence no other machine may take our task (would require
            # lock) and re-process it.
            if status or delete_if_not:
                self.redis.delete(self.key)

        return status

    def options(self):
        """
        Get the options from other _compatible_ (see below) Redis tasks, and
        plan the list of tasks in self.notify_keys[] that we will notify in
        commit().

        We don't merge tasks that have a different 'devel' parameter.  We
        wouldn't be able to tell what sub-tasks are to be created in/out the
        devel subdirectory.

        Similarly, we don't group tasks that have different 'appstream' value.
        That's because normally (not-grouped situation) the final state of
        repository would be order dependent => e.g. if build_A requires
        appstream metadata, and build_B doesn't, the B appstream metadata would
        be added only if build_A was processed after build_B (not vice versa).
        This problem is something we don't want to solve at options() level, and
        we want rather let two concurrent processes in race (it requires at
        least one more createrepo run, but the "appstream" flag shouldn't change
        frequently anyway).
        """
        add = set()
        delete = set()
        rpms_to_remove = set()
        full = False

        if self.noop:
            return (full, add, delete, rpms_to_remove)

        for key in self.redis.keys(self.key_pattern):
            assert key != self.key

            task_dict = self.redis.hgetall(key)
            if not task_dict:
                # TOCTOU: The key might no longer exist in DB, even though
                # _our process_ holds the lock!  See the
                # 'self.redis.delete(self.key)' above in check_processed(); the
                # _original process_ removes key without locking, if _third_
                # process commit()ted the status=succeeded.
                # Prior fixing #3770/#3777 we tried to process these deleted
                # tasks and ended up with KeyError on task_dict["task"] below.
                self.log.info("Key %s already processed (by other process), "
                              "removed by the original process (without lock) "
                              "since we listed it.", key)
                continue

            if task_dict.get("status") is not None:
                # skip processed tasks
                self.log.info("Key %s already processed, skip", key)
                continue

            task_opts = json.loads(task_dict["task"])

            skip = False
            for attr in ["devel", "appstream"]:
                our_value = getattr(self, attr)
                if task_opts[attr] != our_value:
                    self.log.info("'%s' attribute doesn't match: %s/%s",
                                  attr, task_opts[attr], our_value)
                    skip = True
                    break

            if skip:
                continue

            # we can process this task!
            self.notify_keys.append(key)

            # inherit "full" request from others
            if task_opts["full"]:
                full = True
                add = set()

            # append "add" tasks, if that makes sense
            if not full:
                add.update(task_opts["add"])

            # always process the delete requests
            delete.update(task_opts["delete"])

            rpms_to_remove.update(task_opts["rpms_to_remove"])

            if len(self.notify_keys) >= MAX_IN_BATCH - 1:  # one is ours!
                self.log.info("Batch copr-repo limit %s reached, skip the rest",
                              MAX_IN_BATCH)
                break

        return (full, add, delete, rpms_to_remove)

    def commit(self):
        """
        Report that we processed other createrepo requests.  We don't report
        about failures, we rather kindly let the responsible processes to re-try
        the createrepo tasks.  Requires lock!
        """
        if self.noop:
            return

        for key in self.notify_keys:
            self.log.info("Notifying %s that we succeeded", key)
            self.redis.hset(key, "status", "success")
